#  Copyright 2013 Mirantis Inc.
#  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import re
import stat

from cassandra.auth import PlainTextAuthProvider
from cassandra.cluster import Cluster
from cassandra.cluster import NoHostAvailable
from cassandra import OperationTimedOut
from cassandra.policies import ConstantReconnectionPolicy
from oslo_log import log as logging
from oslo_utils import netutils

from trove.common import cfg
from trove.common.db.cassandra import models
from trove.common import exception
from trove.common.i18n import _
from trove.common import instance as rd_instance
from trove.common.stream_codecs import IniCodec
from trove.common.stream_codecs import PropertiesCodec
from trove.common.stream_codecs import SafeYamlCodec
from trove.common.stream_codecs import XmlCodec
from trove.common import utils
from trove.guestagent.common.configuration import ConfigurationManager
from trove.guestagent.common.configuration import OneFileOverrideStrategy
from trove.guestagent.common import guestagent_utils
from trove.guestagent.common import operating_system
from trove.guestagent.common.operating_system import FileMode
from trove.guestagent.datastore import service
from trove.guestagent import pkg


LOG = logging.getLogger(__name__)
CONF = cfg.CONF

packager = pkg.Package()


class CassandraApp(object):
    """Prepares DBaaS on a Guest container."""

    _ADMIN_USER = 'os_admin'

    _CONF_AUTH_SEC = 'authentication'
    _CONF_USR_KEY = 'username'
    _CONF_PWD_KEY = 'password'
    _CONF_DIR_MODS = stat.S_IRWXU
    _CONF_FILE_MODS = stat.S_IRUSR

    CASSANDRA_CONF_FILE = "cassandra.yaml"
    CASSANDRA_TOPOLOGY_FILE = 'cassandra-rackdc.properties'
    CASSANDRA_LOGBACK_FILE = "logback.xml"

    _TOPOLOGY_CODEC = PropertiesCodec(
        delimiter='=', unpack_singletons=True, string_mappings={
            'true': True, 'false': False})

    CASSANDRA_KILL_CMD = "sudo killall java  || true"

    def __init__(self):
        self.state_change_wait_time = CONF.state_change_wait_time
        self.status = CassandraAppStatus(self.get_current_superuser())

        revision_dir = guestagent_utils.build_file_path(
            os.path.dirname(self.cassandra_conf),
            ConfigurationManager.DEFAULT_STRATEGY_OVERRIDES_SUB_DIR)
        self.configuration_manager = ConfigurationManager(
            self.cassandra_conf,
            self.cassandra_owner, self.cassandra_owner,
            SafeYamlCodec(default_flow_style=False), requires_root=True,
            override_strategy=OneFileOverrideStrategy(revision_dir))

        lb_revision_dir = guestagent_utils.build_file_path(
            os.path.dirname(self.cassandra_logback), 'logback-overrides')
        self.logback_conf_manager = ConfigurationManager(
            self.cassandra_logback,
            self.cassandra_owner, self.cassandra_owner,
            XmlCodec(), requires_root=True,
            override_strategy=OneFileOverrideStrategy(lb_revision_dir))

    @property
    def service_candidates(self):
        return ['cassandra']

    @property
    def cassandra_conf_dir(self):
        return {
            operating_system.REDHAT: "/etc/cassandra/default.conf/",
            operating_system.DEBIAN: "/etc/cassandra/",
            operating_system.SUSE: "/etc/cassandra/default.conf/"
        }[operating_system.get_os()]

    @property
    def cassandra_conf(self):
        return guestagent_utils.build_file_path(self.cassandra_conf_dir,
                                                self.CASSANDRA_CONF_FILE)

    @property
    def cassandra_topology(self):
        return guestagent_utils.build_file_path(self.cassandra_conf_dir,
                                                self.CASSANDRA_TOPOLOGY_FILE)

    @property
    def cassandra_owner(self):
        return 'cassandra'

    @property
    def cassandra_data_dir(self):
        return guestagent_utils.build_file_path(
            self.cassandra_working_dir, 'data')

    @property
    def cassandra_working_dir(self):
        return "/var/lib/cassandra"

    @property
    def cassandra_system_log_file(self):
        return guestagent_utils.build_file_path(
            self.cassandra_log_dir, 'system', 'log')

    @property
    def cassandra_log_dir(self):
        return "/var/log/cassandra"

    @property
    def cassandra_logback(self):
        return guestagent_utils.build_file_path(self.cassandra_conf_dir,
                                                self.CASSANDRA_LOGBACK_FILE)

    @property
    def default_superuser_password(self):
        return "cassandra"

    @property
    def default_superuser_pwd_hash(self):
        # Default 'salted_hash' value for 'cassandra' user on Cassandra 2.1.
        return "$2a$10$wPEVuXBU7WE2Uwzqq3t19ObRJyoKztzC/Doyfr0VtDmVXC4GDAV3e"

    @property
    def cqlsh_conf_path(self):
        return "~/.cassandra/cqlshrc"

    def build_admin(self):
        return CassandraAdmin(self.get_current_superuser())

    def install_if_needed(self, packages):
        """Prepare the guest machine with a Cassandra server installation."""
        LOG.info("Preparing Guest as a Cassandra Server")
        if not packager.pkg_is_installed(packages):
            self._install_db(packages)
        LOG.debug("Cassandra install_if_needed complete")

    def init_storage_structure(self, mount_point):
        try:
            operating_system.create_directory(mount_point, as_root=True)
        except exception.ProcessExecutionError:
            LOG.exception("Error while initiating storage structure.")

    def start_db(self, update_db=False, enable_on_boot=True):
        self.status.start_db_service(
            self.service_candidates, self.state_change_wait_time,
            enable_on_boot=enable_on_boot, update_db=update_db)

    def stop_db(self, update_db=False, do_not_start_on_reboot=False):
        self.status.stop_db_service(
            self.service_candidates, self.state_change_wait_time,
            disable_on_boot=do_not_start_on_reboot, update_db=update_db)

    def restart(self):
        self.status.restart_db_service(
            self.service_candidates, self.state_change_wait_time)

    def _install_db(self, packages):
        """Install Cassandra server"""
        LOG.debug("Installing Cassandra server.")
        packager.pkg_install(packages, None, 10000)
        LOG.debug("Finished installing Cassandra server")

    def _remove_system_tables(self):
        """
        Clean up the system keyspace.

        System tables are initialized on the first boot.
        They store certain properties, such as 'cluster_name',
        that cannot be easily changed once afterwards.
        The system keyspace needs to be cleaned up first. The
        tables will be regenerated on the next startup.
        Make sure to also cleanup the commitlog and caches to avoid
        startup errors due to inconsistencies.

        The service should not be running at this point.
        """
        if self.status.is_running:
            raise RuntimeError(_("Cannot remove system tables. "
                                 "The service is still running."))

        LOG.info('Removing existing system tables.')
        system_keyspace_dir = guestagent_utils.build_file_path(
            self.cassandra_data_dir, 'system')
        commitlog_file = guestagent_utils.build_file_path(
            self.cassandra_working_dir, 'commitlog')
        chaches_dir = guestagent_utils.build_file_path(
            self.cassandra_working_dir, 'saved_caches')

        operating_system.remove(system_keyspace_dir,
                                force=True, recursive=True, as_root=True)
        operating_system.remove(commitlog_file,
                                force=True, recursive=True, as_root=True)
        operating_system.remove(chaches_dir,
                                force=True, recursive=True, as_root=True)

        operating_system.create_directory(
            system_keyspace_dir,
            user=self.cassandra_owner, group=self.cassandra_owner,
            force=True, as_root=True)
        operating_system.create_directory(
            commitlog_file,
            user=self.cassandra_owner, group=self.cassandra_owner,
            force=True, as_root=True)
        operating_system.create_directory(
            chaches_dir,
            user=self.cassandra_owner, group=self.cassandra_owner,
            force=True, as_root=True)

    def _apply_post_restore_updates(self, backup_info):
        """The service should not be running at this point.

        The restored database files carry some properties over from the
        original instance that need to be updated with appropriate
        values for the new instance.
        These include:

            - Reset the 'cluster_name' property to match the new unique
              ID of this instance.
              This is to ensure that the restored instance is a part of a new
              single-node cluster rather than forming a one with the
              original node.
            - Reset the administrator's password.
              The original password from the parent instance may be
              compromised or long lost.

        A general procedure is:
            - update the configuration property with the current value
              so that the service can start up
            - reset the superuser password
            - restart the service
            - change the cluster name
            - restart the service

        :seealso: _reset_admin_password
        :seealso: change_cluster_name
        """

        if self.status.is_running:
            raise RuntimeError(_("Cannot reset the cluster name. "
                                 "The service is still running."))

        LOG.debug("Applying post-restore updates to the database.")

        try:
            # Change the 'cluster_name' property to the current in-database
            # value so that the database can start up.
            self._update_cluster_name_property(backup_info['instance_id'])

            # Reset the superuser password so that we can log-in.
            self._reset_admin_password()

            # Start the database and update the 'cluster_name' to the
            # new value.
            self.start_db(update_db=False)
            self.change_cluster_name(CONF.guest_id)
        finally:
            self.stop_db()  # Always restore the initial state of the service.

    def cluster_secure(self, password):
        return self.secure(password=password).serialize()

    def secure(self, update_user=None, password=None):
        """Configure the Trove administrative user.
        Update an existing user if given.
        Create a new one using the default database credentials
        otherwise and drop the built-in user when finished.
        """
        LOG.info('Configuring Trove superuser.')

        if password is None:
            password = utils.generate_random_password()

        admin_username = update_user.name if update_user else self._ADMIN_USER
        os_admin = models.CassandraUser(admin_username, password)

        if update_user:
            CassandraAdmin(update_user).alter_user_password(os_admin)
        else:
            cassandra = models.CassandraUser(
                models.CassandraUser.root_username,
                self.default_superuser_password)
            CassandraAdmin(cassandra)._create_superuser(os_admin)
            CassandraAdmin(os_admin).drop_user(cassandra)

        self._update_admin_credentials(os_admin)

        return os_admin

    def _update_admin_credentials(self, user):
        self.__create_cqlsh_config({self._CONF_AUTH_SEC:
                                    {self._CONF_USR_KEY: user.name,
                                     self._CONF_PWD_KEY: user.password}})

        # Update the internal status with the new user.
        self.status = CassandraAppStatus(user)

    def store_admin_credentials(self, admin_credentials):
        user = models.CassandraUser.deserialize(admin_credentials)
        self._update_admin_credentials(user)

    def get_admin_credentials(self):
        return self.get_current_superuser().serialize()

    def _reset_admin_password(self):
        """
        Reset the password of the Trove's administrative superuser.

        The service should not be running at this point.

        A general password reset procedure is:
            - disable user authentication and remote access
            - restart the service
            - update the password in the 'system_auth.credentials' table
            - re-enable authentication and make the host reachable
            - restart the service
        """
        if self.status.is_running:
            raise RuntimeError(_("Cannot reset the administrative password. "
                                 "The service is still running."))

        try:
            # Disable automatic startup in case the node goes down before
            # we have the superuser secured.
            operating_system.disable_service_on_boot(self.service_candidates)

            self.__disable_remote_access()
            self.__disable_authentication()

            # We now start up the service and immediately re-enable
            # authentication in the configuration file (takes effect after
            # restart).
            # Then we reset the superuser password to its default value
            # and restart the service to get user functions back.
            self.start_db(update_db=False, enable_on_boot=False)
            self.__enable_authentication()
            os_admin = self.__reset_user_password_to_default(self._ADMIN_USER)
            self.status = CassandraAppStatus(os_admin)
            self.restart()

            # Now change the administrative password to a new secret value.
            self.secure(update_user=os_admin)
        finally:
            self.stop_db()  # Always restore the initial state of the service.

        # At this point, we should have a secured database with new Trove-only
        # superuser password.
        # Proceed to re-enable remote access and automatic startup.
        self.__enable_remote_access()
        operating_system.enable_service_on_boot(self.service_candidates)

    def __reset_user_password_to_default(self, username):
        LOG.debug("Resetting the password of user '%(user)s' to '%(pw)s'.",
                  {'user': username, 'pw': self.default_superuser_password})

        user = models.CassandraUser(username, self.default_superuser_password)
        with CassandraLocalhostConnection(user) as client:
            client.execute(
                "UPDATE system_auth.credentials SET salted_hash=%s "
                "WHERE username='{}';", (user.name,),
                (self.default_superuser_pwd_hash,))

            return user

    def change_cluster_name(self, cluster_name):
        """Change the 'cluster_name' property of an exesting running instance.
        Cluster name is stored in the database and is required to match the
        configuration value. Cassandra fails to start otherwise.
        """

        if not self.status.is_running:
            raise RuntimeError(_("Cannot change the cluster name. "
                                 "The service is not running."))

        LOG.debug("Changing the cluster name to '%s'.", cluster_name)

        # Update the in-database value.
        self.__reset_cluster_name(cluster_name)

        # Update the configuration property.
        self._update_cluster_name_property(cluster_name)

        self.restart()

    def __reset_cluster_name(self, cluster_name):
        # Reset the in-database value stored locally on this node.
        current_superuser = self.get_current_superuser()
        with CassandraLocalhostConnection(current_superuser) as client:
            client.execute(
                "UPDATE system.local SET cluster_name = '{}' "
                "WHERE key='local';", (cluster_name,))

        # Newer version of Cassandra require a flush to ensure the changes
        # to the local system keyspace persist.
        self.flush_tables('system', 'local')

    def __create_cqlsh_config(self, sections):
        config_path = self._get_cqlsh_conf_path()
        config_dir = os.path.dirname(config_path)
        if not os.path.exists(config_dir):
            os.mkdir(config_dir, self._CONF_DIR_MODS)
        else:
            os.chmod(config_dir, self._CONF_DIR_MODS)
        operating_system.write_file(config_path, sections, codec=IniCodec())
        os.chmod(config_path, self._CONF_FILE_MODS)

    def get_current_superuser(self):
        """
        Build the Trove superuser.
        Use the stored credentials.
        If not available fall back to the defaults.
        """
        if self.has_user_config():
            return self._load_current_superuser()

        LOG.warning(
            "Trove administrative user has not been configured yet. "
            "Using the built-in default: %s",
            models.CassandraUser.root_username)
        return models.CassandraUser(models.CassandraUser.root_username,
                                    self.default_superuser_password)

    def has_user_config(self):
        """
        Return TRUE if there is a client configuration file available
        on the guest.
        """
        return os.path.exists(self._get_cqlsh_conf_path())

    def _load_current_superuser(self):
        config = operating_system.read_file(self._get_cqlsh_conf_path(),
                                            codec=IniCodec())
        return models.CassandraUser(
            config[self._CONF_AUTH_SEC][self._CONF_USR_KEY],
            config[self._CONF_AUTH_SEC][self._CONF_PWD_KEY]
        )

    def apply_initial_guestagent_configuration(self, cluster_name=None):
        """Update guestagent-controlled configuration properties.
        These changes to the default template are necessary in order to make
        the database service bootable and accessible in the guestagent context.

        :param cluster_name:  The 'cluster_name' configuration property.
                              Use the unique guest id by default.
        :type cluster_name:   string
        """
        self.configuration_manager.apply_system_override(
            {'data_file_directories': [self.cassandra_data_dir]})
        self._make_host_reachable()
        self._update_cluster_name_property(cluster_name or CONF.guest_id)
        # A single-node instance may use the SimpleSnitch
        # (keyspaces use SimpleStrategy).
        # A network-aware snitch has to be used otherwise.
        if cluster_name is None:
            updates = {'endpoint_snitch': 'SimpleSnitch'}
        else:
            updates = {'endpoint_snitch': 'GossipingPropertyFileSnitch'}
        self.configuration_manager.apply_system_override(updates)

    def _make_host_reachable(self):
        """
        Some of these settings may be overridden by user defined
        configuration groups.

        authenticator and authorizer
            - Necessary to enable users and permissions.
        rpc_address - Enable remote connections on all interfaces.
        broadcast_rpc_address - RPC address to broadcast to drivers and
                                other clients. Must be set if
                                rpc_address = 0.0.0.0 and can never be
                                0.0.0.0 itself.
        listen_address - The address on which the node communicates with
                         other nodes. Can never be 0.0.0.0.
        seed_provider - A list of discovery contact points.
        """
        self.__enable_authentication()
        self.__enable_remote_access()

    def __enable_remote_access(self):
        updates = {
            'rpc_address': "0.0.0.0",
            'broadcast_rpc_address': netutils.get_my_ipv4(),
            'listen_address': netutils.get_my_ipv4(),
            'seed_provider': {'parameters':
                              [{'seeds': netutils.get_my_ipv4()}]
                              }
        }

        self.configuration_manager.apply_system_override(updates)

    def __disable_remote_access(self):
        updates = {
            'rpc_address': "127.0.0.1",
            'listen_address': '127.0.0.1',
            'seed_provider': {'parameters':
                              [{'seeds': '127.0.0.1'}]
                              }
        }

        self.configuration_manager.apply_system_override(updates)

    def __enable_authentication(self):
        updates = {
            'authenticator': 'org.apache.cassandra.auth.PasswordAuthenticator',
            'authorizer': 'org.apache.cassandra.auth.CassandraAuthorizer'
        }

        self.configuration_manager.apply_system_override(updates)

    def __disable_authentication(self):
        updates = {
            'authenticator': 'org.apache.cassandra.auth.AllowAllAuthenticator',
            'authorizer': 'org.apache.cassandra.auth.AllowAllAuthorizer'
        }

        self.configuration_manager.apply_system_override(updates)

    def _update_cluster_name_property(self, name):
        """This 'cluster_name' property prevents nodes from one
        logical cluster from talking to another.
        All nodes in a cluster must have the same value.
        """
        self.configuration_manager.apply_system_override({'cluster_name':
                                                          name})

    def update_overrides(self, context, overrides, remove=False):
        if overrides:
            self.configuration_manager.apply_user_override(overrides)

    def remove_overrides(self):
        self.configuration_manager.remove_user_override()

    def write_cluster_topology(self, data_center, rack, prefer_local=True):
        LOG.info('Saving Cassandra cluster topology configuration.')

        config = {'dc': data_center,
                  'rack': rack,
                  'prefer_local': prefer_local}

        operating_system.write_file(self.cassandra_topology, config,
                                    codec=self._TOPOLOGY_CODEC, as_root=True)
        operating_system.chown(
            self.cassandra_topology,
            self.cassandra_owner, self.cassandra_owner, as_root=True)
        operating_system.chmod(
            self.cassandra_topology, FileMode.ADD_READ_ALL, as_root=True)

    def start_db_with_conf_changes(self, config_contents):
        LOG.debug("Starting database with configuration changes.")
        if self.status.is_running:
            raise RuntimeError(_("The service is still running."))

        self.configuration_manager.save_configuration(config_contents)
        # The configuration template has to be updated with
        # guestagent-controlled settings.
        self.apply_initial_guestagent_configuration()
        self.start_db(True)

    def reset_configuration(self, configuration):
        LOG.debug("Resetting configuration.")
        config_contents = configuration['config_contents']
        self.configuration_manager.save_configuration(config_contents)

    def _get_cqlsh_conf_path(self):
        return os.path.expanduser(self.cqlsh_conf_path)

    def get_data_center(self):
        config = operating_system.read_file(self.cassandra_topology,
                                            codec=self._TOPOLOGY_CODEC)
        return config['dc']

    def get_rack(self):
        config = operating_system.read_file(self.cassandra_topology,
                                            codec=self._TOPOLOGY_CODEC)
        return config['rack']

    def set_seeds(self, seeds):
        LOG.debug("Setting seed nodes: %s", seeds)
        updates = {
            'seed_provider': {'parameters':
                              [{'seeds': ','.join(seeds)}]
                              }
        }

        self.configuration_manager.apply_system_override(updates)

    def get_seeds(self):
        """Return a list of seed node IPs if any.

        The seed IPs are stored as a comma-separated string in the
        seed-provider parameters:
        [{'class_name': '<name>', 'parameters': [{'seeds': '<ip>,<ip>'}, ...]}]
        """

        def find_first(key, dict_list):
            for item in dict_list:
                if key in item:
                    return item[key]
            return []

        sp_property = self.configuration_manager.get_value('seed_provider', [])
        seeds_str = find_first('seeds', find_first('parameters', sp_property))
        return seeds_str.split(',') if seeds_str else []

    def set_auto_bootstrap(self, enabled):
        """Auto-bootstrap makes new (non-seed) nodes automatically migrate the
        right data to themselves.
        The feature has to be turned OFF when initializing a fresh cluster
        without data.
        It must be turned back ON once the cluster is initialized.
        """
        LOG.debug("Setting auto-bootstrapping: %s", enabled)
        updates = {'auto_bootstrap': enabled}
        self.configuration_manager.apply_system_override(updates)

    def node_cleanup_begin(self):
        """Suspend periodic status updates and mark the instance busy
        throughout the operation.
        """
        self.status.begin_restart()
        self.status.set_status(rd_instance.ServiceStatuses.BLOCKED)

    def node_cleanup(self):
        """Cassandra does not automatically remove data from nodes that
        lose part of their partition range to a newly added node.
        Cleans up keyspaces and partition keys no longer belonging to the node.

        Do not treat cleanup failures as fatal. Resume the heartbeat after
        finishing and let it signal the true state of the instance to the
        caller.
        """
        LOG.debug("Running node cleanup.")
        # nodetool -h <HOST> -p <PORT> -u <USER> -pw <PASSWORD> cleanup
        try:
            self._run_nodetool_command('cleanup')
            self.status.set_status(rd_instance.ServiceStatuses.RUNNING)
        except Exception:
            LOG.exception("The node failed to complete its cleanup.")
        finally:
            self.status.end_restart()

    def node_decommission(self):
        """Causes a live node to decommission itself,
        streaming its data to the next node on the ring.

        Shutdown the database after successfully finishing the operation,
        or leave the node in a failed state otherwise.

        Suspend periodic status updates, so that the caller can poll for the
        database shutdown.
        """
        LOG.debug("Decommissioning the node.")
        # nodetool -h <HOST> -p <PORT> -u <USER> -pw <PASSWORD> decommission
        self.status.begin_restart()
        try:
            self._run_nodetool_command('decommission')
        except Exception:
            LOG.exception("The node failed to decommission itself.")
            self.status.set_status(rd_instance.ServiceStatuses.FAILED)
            return
        finally:
            # Cassandra connections have ability to automatically discover and
            # fallback to other cluster nodes whenever a node goes down.
            # Reset the status after decomissioning to ensure the heartbeat
            # connection talks to this node only.
            self.status = CassandraAppStatus(self.get_current_superuser())

        try:
            self.stop_db(update_db=True, do_not_start_on_reboot=True)
        finally:
            self.status.end_restart()

    def flush_tables(self, keyspace, *tables):
        """Flushes one or more tables from the memtable.
        """
        LOG.debug("Flushing tables.")
        # nodetool -h <HOST> -p <PORT> -u <USER> -pw <PASSWORD> flush --
        # <keyspace> ( <table> ... )
        self._run_nodetool_command('flush', keyspace, *tables)

    def set_logging_level(self, log_level):
        """Set the log Cassandra's system log verbosity level.
        """
        # Apply the change at runtime.
        self._run_nodetool_command('setlogginglevel', 'root', log_level)

        # Persist the change.
        self.logback_conf_manager.apply_system_override(
            {'configuration': {'root': {'@level': log_level}}})

    def drain(self):
        """Drains Cassandra node so that it can upgraded safely.
        """
        LOG.debug("Draining node.")
        self._run_nodetool_command('drain')

    def upgrade_sstables(self):
        """Upgrades sstables to match new datastore version.
        """
        LOG.debug("Upgrading sstables.")
        self._run_nodetool_command('upgradesstables')

    def _run_nodetool_command(self, cmd, *args, **kwargs):
        """Execute a nodetool command on this node.
        """
        return utils.execute('nodetool', '-h', 'localhost',
                             cmd, *args, **kwargs)

    def enable_root(self, root_password=None):
        """Cassandra's 'root' user is called 'cassandra'.
        Create a new superuser if it does not exist and grant it full
        superuser-level access to all keyspaces.
        """
        cassandra = models.CassandraUser.root(password=root_password)
        admin = self.build_admin()
        if self.is_root_enabled():
            admin.alter_user_password(cassandra)
        else:
            admin._create_superuser(cassandra)

        return cassandra.serialize()

    def is_root_enabled(self):
        """The Trove administrative user ('os_admin') should normally be the
        only superuser in the system.
        """
        found = self.build_admin().list_superusers()
        return len([user for user in found
                    if user.name != self._ADMIN_USER]) > 0


class CassandraAppStatus(service.BaseDbStatus):

    def __init__(self, superuser):
        """
        :param superuser:        User account the Status uses for connecting
                                 to the database.
        :type superuser:         CassandraUser
        """
        super(CassandraAppStatus, self).__init__()
        self.__user = superuser
        self.__client = None

    @property
    def client(self):
        if self.__client is None:
            self.__client = CassandraLocalhostConnection(self.__user)
        return self.__client

    def _get_actual_db_status(self):
        try:
            if self.client.local_node_is_up():
                return rd_instance.ServiceStatuses.RUNNING
        except NoHostAvailable:
            return rd_instance.ServiceStatuses.SHUTDOWN
        except Exception:
            LOG.exception("Error getting Cassandra status.")

        return rd_instance.ServiceStatuses.SHUTDOWN

    def cleanup_stalled_db_services(self):
        utils.execute_with_timeout(CassandraApp.CASSANDRA_KILL_CMD, shell=True)


class CassandraAdmin(object):
    """Handles administrative tasks on the Cassandra database.

    In Cassandra only SUPERUSERS can create other users and grant permissions
    to database resources. Trove uses the 'cassandra' superuser to perform its
    administrative tasks.

    The users it creates are all 'normal' (NOSUPERUSER) accounts.
    The permissions it can grant are also limited to non-superuser operations.
    This is to prevent anybody from creating a new superuser via the Trove API.
    """

    # Non-superuser grant modifiers.
    __NO_SUPERUSER_MODIFIERS = ('ALTER', 'CREATE', 'DROP', 'MODIFY', 'SELECT')

    _KS_NAME_REGEX = re.compile('^<keyspace (.+)>$')

    def __init__(self, user):
        self.__admin_user = user
        self.__client = None

    @property
    def client(self):
        if self.__client is None:
            self.__client = CassandraLocalhostConnection(self.__admin_user)
        return self.__client

    def create_user(self, context, users):
        """
        Create new non-superuser accounts.
        New users are by default granted full access to all database resources.
        """
        for item in users:
            self._create_user_and_grant(self.client,
                                        self._deserialize_user(item))

    def _create_user_and_grant(self, client, user):
        """
        Create new non-superuser account and grant it full access to its
        databases.
        """
        self._create_user(client, user)
        for db in user.databases:
            self._grant_full_access_on_keyspace(
                client, self._deserialize_keyspace(db), user)

    def _create_user(self, client, user):
        # Create only NOSUPERUSER accounts here.
        LOG.debug("Creating a new user '%s'.", user.name)
        client.execute("CREATE USER '{}' WITH PASSWORD %s NOSUPERUSER;",
                       (user.name,), (user.password,))

    def _create_superuser(self, user):
        """Create a new superuser account and grant it full superuser-level
        access to all keyspaces.
        """
        LOG.debug("Creating a new superuser '%s'.", user.name)
        self.client.execute("CREATE USER '{}' WITH PASSWORD %s SUPERUSER;",
                            (user.name,), (user.password,))
        self.client.execute(
            "GRANT ALL PERMISSIONS ON ALL KEYSPACES TO '{}';", (user.name,))

    def delete_user(self, context, user):
        self.drop_user(self._deserialize_user(user))

    def drop_user(self, user):
        self._drop_user(self.client, user)

    def _drop_user(self, client, user):
        LOG.debug("Deleting user '%s'.", user.name)
        client.execute("DROP USER '{}';", (user.name, ))

    def get_user(self, context, username, hostname):
        user = self._find_user(self.client, username)
        return user.serialize() if user is not None else None

    def _find_user(self, client, username):
        """
        Lookup a user with a given username.
        Omit user names on the ignore list.
        Return a new Cassandra user instance or None if no match is found.
        """
        return next((user for user in self._get_listed_users(client)
                     if user.name == username), None)

    def list_users(self, context, limit=None, marker=None,
                   include_marker=False):
        """
        List all non-superuser accounts. Omit names on the ignored list.
        Return an empty set if None.
        """
        return guestagent_utils.serialize_list(
            self._get_listed_users(self.client),
            limit=limit, marker=marker, include_marker=include_marker)

    def _get_listed_users(self, client):
        """
        Return a set of unique user instances.
        Omit user names on the ignore list.
        """
        return self._get_users(
            client, lambda user: user.name not in self.ignore_users)

    def _get_users(self, client, matcher=None):
        """
        :param matcher                Filter expression.
        :type matcher                 callable
        """
        acl = self._get_acl(client)
        return {self._build_user(user.name, acl)
                for user in client.execute("LIST USERS;")
                if not matcher or matcher(user)}

    def _load_user(self, client, username, check_reserved=True):
        if check_reserved:
            models.CassandraUser(username).check_reserved()

        acl = self._get_acl(client, username=username)
        return self._build_user(username, acl)

    def _build_user(self, username, acl):
        user = models.CassandraUser(username)
        for ks, permissions in acl.get(username, {}).items():
            if permissions:
                user.databases.append(models.CassandraSchema(ks).serialize())
        return user

    def _get_acl(self, client, username=None):
        """Return the ACL for a database user.
        Return ACLs for all users if no particular username is specified.

        The ACL has the following format:
        {username #1:
            {keyspace #1: {access mod(s)...},
             keyspace #2: {...}},
         username #2:
            {keyspace #1: {...},
             keyspace #3: {...}}
        }
        """

        def build_list_query(username):
            query_tokens = ["LIST ALL PERMISSIONS"]
            if username:
                query_tokens.extend(["OF", "'%s'" % username])
            query_tokens.append("NORECURSIVE;")
            return ' '.join(query_tokens)

        def parse_keyspace_name(resource):
            """Parse a keyspace name from a resource string.
            The resource string has the following form:
                <object name>
            where 'object' is one of the database objects (keyspace, table...).
            Return the name as a singleton set. Return an empty set if no match
            is found.
            """
            match = self._KS_NAME_REGEX.match(resource)
            if match:
                return {match.group(1)}
            return {}

        def update_acl(username, keyspace, permission, acl):
            permissions = acl.get(username, {}).get(keyspace)
            if permissions is None:
                guestagent_utils.update_dict({user: {keyspace: {permission}}},
                                             acl)
            else:
                permissions.add(permission)

        all_keyspace_names = None
        acl = dict()
        for item in client.execute(build_list_query(username)):
            user = item.username
            resource = item.resource
            permission = item.permission
            if user and resource and permission:
                if resource == '<all keyspaces>':
                    # Cache the full keyspace list to improve performance and
                    # ensure consistent results for all users.
                    if all_keyspace_names is None:
                        all_keyspace_names = {
                            item.name
                            for item in self._get_available_keyspaces(client)
                        }
                    keyspaces = all_keyspace_names
                else:
                    keyspaces = parse_keyspace_name(resource)

                for keyspace in keyspaces:
                    update_acl(user, keyspace, permission, acl)

        return acl

    def list_superusers(self):
        """List all system users existing in the database."""
        return self._get_users(self.client, lambda user: user.super)

    def grant_access(self, context, username, hostname, databases):
        """
        Grant full access on keyspaces to a given username.
        """
        user = models.CassandraUser(username)
        for db in databases:
            self._grant_full_access_on_keyspace(
                self.client, models.CassandraSchema(db), user)

    def revoke_access(self, context, username, hostname, database):
        """
        Revoke all permissions on any database resources from a given username.
        """
        user = models.CassandraUser(username)
        self._revoke_all_access_on_keyspace(
            self.client, models.CassandraSchema(database), user)

    def _grant_full_access_on_keyspace(self, client, keyspace, user,
                                       check_reserved=True):
        """
        Grant all non-superuser permissions on a keyspace to a given user.
        """
        if check_reserved:
            user.check_reserved()
            keyspace.check_reserved()

        for access in self.__NO_SUPERUSER_MODIFIERS:
            self._grant_permission_on_keyspace(client, access, keyspace, user)

    def _grant_permission_on_keyspace(self, client, modifier, keyspace, user):
        """
        Grant a non-superuser permission on a keyspace to a given user.
        Raise an exception if the caller attempts to grant a superuser access.
        """
        LOG.debug("Granting '%(mod)s' access on '%(keyspace_name)s' to "
                  "user '%(user)s'.",
                  {'mod': modifier, 'keyspace_name': keyspace.name,
                   'user': user.name})
        if modifier in self.__NO_SUPERUSER_MODIFIERS:
            client.execute("GRANT {} ON KEYSPACE \"{}\" TO '{}';",
                           (modifier, keyspace.name, user.name))
        else:
            raise exception.UnprocessableEntity(
                "Invalid permission modifier (%s). Allowed values are: '%s'"
                % (modifier, ', '.join(self.__NO_SUPERUSER_MODIFIERS)))

    def _revoke_all_access_on_keyspace(self, client, keyspace, user,
                                       check_reserved=True):
        if check_reserved:
            user.check_reserved()
            keyspace.check_reserved()

        LOG.debug("Revoking all permissions on '%(keyspace_name)s' "
                  "from user '%(user)s'.", {'keyspace_name': keyspace.name,
                                            'user': user.name})
        client.execute("REVOKE ALL PERMISSIONS ON KEYSPACE \"{}\" FROM '{}';",
                       (keyspace.name, user.name))

    def update_attributes(self, context, username, hostname, user_attrs):
        user = self._load_user(self.client, username)
        new_name = user_attrs.get('name')
        new_password = user_attrs.get('password')
        self._update_user(self.client, user, new_name, new_password)

    def _update_user(self, client, user, new_username, new_password):
        """
        Update a user of a given username.
        Updatable attributes include username and password.
        If a new username and password are given a new user with those
        attributes is created and all permissions from the original
        user get transferred to it. The original user is then dropped
        therefore revoking its permissions.
        If only new password is specified the existing user gets altered
        with that password.
        """
        if new_username is not None and user.name != new_username:
            if new_password is not None:
                self._rename_user(client, user, new_username, new_password)
            else:
                raise exception.UnprocessableEntity(
                    _("Updating username requires specifying a password "
                      "as well."))
        elif new_password is not None and user.password != new_password:
            user.password = new_password
            self._alter_user_password(client, user)

    def _rename_user(self, client, user, new_username, new_password):
        """
        Rename a given user also updating its password.
        Transfer the current permissions to the new username.
        Drop the old username therefore revoking its permissions.
        """
        LOG.debug("Renaming user '%(old)s' to '%(new)s'",
                  {'old': user.name, 'new': new_username})
        new_user = models.CassandraUser(new_username, new_password)
        new_user.databases.extend(user.databases)
        self._create_user_and_grant(client, new_user)
        self._drop_user(client, user)

    def alter_user_password(self, user):
        self._alter_user_password(self.client, user)

    def change_passwords(self, context, users):
        for user in users:
            self._alter_user_password(self.client,
                                      self._deserialize_user(user))

    def _alter_user_password(self, client, user):
        LOG.debug("Changing password of user '%s'.", user.name)
        client.execute("ALTER USER '{}' "
                       "WITH PASSWORD %s;", (user.name,), (user.password,))

    def create_database(self, context, databases):
        for item in databases:
            self._create_single_node_keyspace(
                self.client, self._deserialize_keyspace(item))

    def _create_single_node_keyspace(self, client, keyspace):
        """
        Create a single-replica keyspace.

        Cassandra stores replicas on multiple nodes to ensure reliability and
        fault tolerance. All replicas are equally important;
        there is no primary or master.
        A replication strategy determines the nodes where
        replicas are placed. SimpleStrategy is for a single data center only.
        The total number of replicas across the cluster is referred to as the
        replication factor.

        Replication Strategy:
        'SimpleStrategy' is not optimized for multiple data centers.
        'replication_factor' The number of replicas of data on multiple nodes.
                             Required for SimpleStrategy; otherwise, not used.

        Keyspace names are case-insensitive by default.
        To make a name case-sensitive, enclose it in double quotation marks.
        """
        client.execute("CREATE KEYSPACE \"{}\" WITH REPLICATION = "
                       "{{ 'class' : 'SimpleStrategy', "
                       "'replication_factor' : 1 }};", (keyspace.name,))

    def delete_database(self, context, database):
        self._drop_keyspace(self.client,
                            self._deserialize_keyspace(database))

    def _drop_keyspace(self, client, keyspace):
        LOG.debug("Dropping keyspace '%s'.", keyspace.name)
        client.execute("DROP KEYSPACE \"{}\";", (keyspace.name,))

    def list_databases(self, context, limit=None, marker=None,
                       include_marker=False):
        return guestagent_utils.serialize_list(
            self._get_available_keyspaces(self.client),
            limit=limit, marker=marker, include_marker=include_marker)

    def _get_available_keyspaces(self, client):
        """
        Return a set of unique keyspace instances.
        Omit keyspace names on the ignore list.
        """
        return {models.CassandraSchema(db.keyspace_name)
                for db in client.execute("SELECT * FROM "
                                         "system.schema_keyspaces;")
                if db.keyspace_name not in self.ignore_dbs}

    def list_access(self, context, username, hostname):
        user = self._find_user(self.client, username)
        if user:
            return user.databases

        raise exception.UserNotFound(uuid=username)

    def _deserialize_keyspace(self, keyspace_dict, check_reserved=True):
        if keyspace_dict:
            db = models.CassandraSchema.deserialize(keyspace_dict)
            if check_reserved:
                db.check_reserved()

            return db

        return None

    def _deserialize_user(self, user_dict, check_reserved=True):
        if user_dict:
            user = models.CassandraUser.deserialize(user_dict)
            if check_reserved:
                user.check_reserved()

            return user

        return None

    @property
    def ignore_users(self):
        return cfg.get_ignored_users()

    @property
    def ignore_dbs(self):
        return cfg.get_ignored_dbs()


class CassandraConnection(object):
    """A wrapper to manage a Cassandra connection."""

    # Cassandra 2.1 only supports protocol versions 3 and lower.
    NATIVE_PROTOCOL_VERSION = 3
    CONNECTION_TIMEOUT_SEC = CONF.agent_call_high_timeout
    RECONNECT_DELAY_SEC = 3

    def __init__(self, contact_points, user):
        self.__user = user
        # A Cluster is initialized with a set of initial contact points.
        # After the driver connects to one of the nodes it will automatically
        # discover the rest.
        # Will connect to '127.0.0.1' if None contact points are given.
        #
        # Set the 'reconnection_policy' so that dead connections recover fast.
        self._cluster = Cluster(
            contact_points=contact_points,
            auth_provider=PlainTextAuthProvider(user.name, user.password),
            protocol_version=self.NATIVE_PROTOCOL_VERSION,
            connect_timeout=self.CONNECTION_TIMEOUT_SEC,
            control_connection_timeout=self.CONNECTION_TIMEOUT_SEC,
            reconnection_policy=ConstantReconnectionPolicy(
                self.RECONNECT_DELAY_SEC, max_attempts=None))
        self.__session = None

        self._connect()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._disconnect()

    def execute(self, query, identifiers=None, data_values=None, timeout=None):
        """
        Execute a query with a given sequence or dict of data values to bind.
        If a sequence is used, '%s' should be used the placeholder for each
        argument. If a dict is used, '%(name)s' style placeholders must
        be used.
        Only data values should be supplied this way. Other items,
        such as keyspaces, table names, and column names should be set
        ahead of time. Use the '{}' style placeholders and
        'identifiers' parameter for those.
        Raise an exception if the operation exceeds the given timeout (sec).
        There is no timeout if set to None.
        Return a set of rows or an empty list if None.
        """
        if self.is_active():
            try:
                rows = self.__session.execute(self.__bind(query, identifiers),
                                              data_values, timeout)
                return rows or []
            except OperationTimedOut:
                LOG.error("Query execution timed out.")
                raise

        LOG.debug("Cannot perform this operation on a closed connection.")
        raise exception.UnprocessableEntity()

    def __bind(self, query, identifiers):
        if identifiers:
            return query.format(*identifiers)
        return query

    def node_is_up(self, host_ip):
        """Test whether the Cassandra node located at the given IP is up.
        """
        for host in self._cluster.metadata.all_hosts():
            if host.address == host_ip:
                return host.is_up
        return False

    def local_node_is_up(self):
        """Test whether Cassandra is up on the localhost.
        """
        return (self.node_is_up('127.0.0.1') or
                self.node_is_up(netutils.get_my_ipv4()))

    def _connect(self):
        if not self._cluster.is_shutdown:
            LOG.debug("Connecting to a Cassandra cluster as '%s'.",
                      self.__user.name)
            if not self.is_active():
                self.__session = self._cluster.connect()
            else:
                LOG.debug("Connection already open.")
            LOG.debug("Connected to cluster: '%s'",
                      self._cluster.metadata.cluster_name)
            for host in self._cluster.metadata.all_hosts():
                LOG.debug("Connected to node: '%(address)s' in rack "
                          "'%(rack)s' at datacenter '%(datacenter)s'",
                          {'address': host.address, 'rack': host.rack,
                           'datacenter': host.datacenter})
        else:
            LOG.debug("Cannot perform this operation on a terminated cluster.")
            raise exception.UnprocessableEntity()

    def _disconnect(self):
        if self.is_active():
            try:
                LOG.debug("Disconnecting from cluster: '%s'",
                          self._cluster.metadata.cluster_name)
                self._cluster.shutdown()
            except Exception:
                LOG.debug("Failed to disconnect from a Cassandra cluster.")

    def is_active(self):
        return self.__session and not self.__session.is_shutdown

    def __del__(self):
        # The connections would survive the parent object's GC.
        # We need to close it explicitly.
        self._disconnect()


class CassandraLocalhostConnection(CassandraConnection):
    """
    A connection to the localhost Cassandra server.
    """

    def __init__(self, user):
        super(CassandraLocalhostConnection, self).__init__(None, user)



# Copyright 2013 Hewlett-Packard Development Company, L.P.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging

from trove.common.strategies.strategy import Strategy

LOG = logging.getLogger(__name__)


def get_restore_strategy(restore_driver, ns=__name__):
    LOG.debug("Getting restore strategy: %s.", restore_driver)
    return Strategy.get_strategy(restore_driver, ns)



#   Copyright (c) 2014 Mirantis, Inc.
#   All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from os import path

from trove.guestagent.common import operating_system
from trove.guestagent import pkg

OS_NAME = operating_system.get_os()

MONGODB_MOUNT_POINT = "/var/lib/mongodb"
MONGO_PID_FILE = '/var/run/mongodb/mongodb.pid'
MONGO_LOG_FILE = '/var/log/mongodb/mongod.log'

CONFIG_CANDIDATES = ["/etc/mongodb.conf", "/etc/mongod.conf"]
MONGO_ADMIN_NAME = 'os_admin'
MONGO_ADMIN_ROLES = [{'db': 'admin', 'role': 'userAdminAnyDatabase'},
                     {'db': 'admin', 'role': 'dbAdminAnyDatabase'},
                     {'db': 'admin', 'role': 'clusterAdmin'},
                     {'db': 'admin', 'role': 'readWriteAnyDatabase'}]
MONGO_ADMIN_CREDS_FILE = path.join(path.expanduser('~'),
                                   '.os_mongo_admin_creds.json')
MONGO_KEY_FILE = '/etc/mongo_key'
MONGOS_SERVICE_CANDIDATES = ["mongos"]
MONGOD_SERVICE_CANDIDATES = ["mongodb", "mongod"]
MONGODB_KILL = "sudo kill %s"
FIND_PID = "ps xaco pid,cmd | awk '/mongo(d|db|s)/ {print $1}'"
TIME_OUT = 1000

MONGO_USER = {operating_system.REDHAT: "mongod",
              operating_system.DEBIAN: "mongodb",
              operating_system.SUSE: "mongod"}[OS_NAME]

PACKAGER = pkg.Package()



# Copyright 2016 IBM Corp
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging
from trove.common.db import models
from trove.common import exception
from trove.common import utils
from trove.guestagent.common import operating_system
from trove.guestagent.datastore.experimental.db2 import service
from trove.guestagent.datastore.experimental.db2 import system
from trove.guestagent.strategies.backup import base

LOG = logging.getLogger(__name__)


class DB2Backup(base.BackupRunner):
    """
     Base class for DB2 backups
    """
    def __init__(self, *args, **kwargs):
        super(DB2Backup, self).__init__(*args, **kwargs)
        self.admin = service.DB2Admin()
        self.databases = self.list_dbnames()

    def list_dbnames(self):
        dbNames = []
        databases, marker = self.admin.list_databases()
        for database in databases:
            mydb = models.DatastoreSchema.deserialize(database)
            dbNames.append(mydb.name)
        return dbNames

    def estimate_backup_size(self):
        """
           Estimating the size of the backup based on the size of the data
           returned from the get_db_size procedure. The size of the
           backup is always going to be smaller than the size of the data.
        """
        try:
            size = 0
            for dbname in self.databases:
                out = service.run_command(system.GET_DB_SIZE % {'dbname':
                                                                dbname})
                size = size + int(out[0])
        except exception.ProcessExecutionError:
            LOG.exception("An error occurred while trying to "
                          "estimate backup size")
        LOG.debug("Estimated size for databases: %d", size)
        return size

    def estimate_log_size(self):
        return 0.0

    def run_backup(self):
        pass

    def execute_backup_cmd(self, backup_command):
        service.create_db2_dir(system.DB2_BACKUP_DIR)
        for dbName in self.databases:
            service.run_command(backup_command % {'dbname': dbName})

    def _run_pre_backup(self):
        """
        Before performing the actual backup we need to make sure that
        there is enough space to store the backups. The backup size
        is the sum of the size of the databases and if it is an online
        backup, the size of the archived logs is also factored in.
        """
        backup_size_bytes = self.estimate_backup_size()
        log_size_bytes = self.estimate_log_size()
        total_backup_size_gb = utils.to_gb(backup_size_bytes + log_size_bytes)
        free_bytes = operating_system.get_bytes_free_on_fs(system.MOUNT_POINT)
        free_gb = utils.to_gb(free_bytes)

        if total_backup_size_gb > free_gb:
            raise exception.InsufficientSpaceForBackup % {
                'backup_size': total_backup_size_gb,
                'free': free_gb
            }
        self.run_backup()

    @property
    def cmd(self):
        cmd = 'sudo tar cPf - ' + system.DB2_BACKUP_DIR
        return cmd + self.zip_cmd + self.encrypt_cmd

    def cleanup(self):
        service.remove_db2_dir(system.DB2_BACKUP_DIR)

    def _run_post_backup(self):
        self.cleanup()


class DB2OnlineBackup(DB2Backup):
    """
    Implementation of Online Backup Strategy for DB2
    using archive logging.
    """
    __strategy_name__ = 'db2onlinebackup'

    def __init__(self, *args, **kwargs):
        super(DB2OnlineBackup, self).__init__(*args, **kwargs)

    def estimate_log_size(self):
        """
        Estimate the log utilization for all databases. The LOG_UTILIZATION
        administrative view returns information about log utilization for the
        connected database. The TOTAL_LOG_USED_KB returns the log utilization
        in KB.
        """
        log_size = 0
        try:
            for dbname in self.databases:
                out = service.run_command(
                    system.LOG_UTILIZATION % {'dbname': dbname})
                log_size = log_size + int(out[0])
            log_size = log_size * 1024
        except exception.ProcessExecutionError:
            LOG.exception("An error occurred while trying to estimate log "
                          "size")
        LOG.debug("Estimated log size for all databases: %d", log_size)
        return log_size

    def run_backup(self):
        try:
            self.execute_backup_cmd(system.ONLINE_BACKUP_DB)
        except exception.ProcessExecutionError:
            LOG.exception("An exception occurred while doing an online "
                          "backup.")
            self.cleanup()
            raise

    def cleanup(self):
        super(DB2OnlineBackup, self).cleanup()
        '''
        After a backup operation, we can delete the archived logs
        from the archived log directory but we do not want to delete
        the directory itself. Since archive logging is enabled for
        all databases, this directory is needed to store archive logs.
        '''
        service.remove_db2_dir(system.DB2_ARCHIVE_LOGS_DIR + "/*")


class DB2OfflineBackup(DB2Backup):
    """
    Implementation of Offline Backup Strategy for DB2 using
    circular logging which is the default.
    """
    __strategy_name__ = 'db2offlinebackup'

    def __init__(self, *args, **kwargs):
        super(DB2OfflineBackup, self).__init__(*args, **kwargs)

    def run_backup(self):
        """Create archival contents in dump dir"""
        try:
            service.run_command(system.QUIESCE_DB2)
            self.execute_backup_cmd(system.OFFLINE_BACKUP_DB)
            service.run_command(system.UNQUIESCE_DB2)
        except exception.ProcessExecutionError:
            LOG.exception("An exception occurred while doing an offline "
                          "backup.")
            self.cleanup()
            raise



# Copyright (c) 2013 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_log import log as logging

from trove.common import exception
from trove.common.i18n import _
from trove.common import instance as rd_instance
from trove.common.notification import EndNotification
from trove.common import utils
from trove.guestagent import backup
from trove.guestagent.common import operating_system
from trove.guestagent.datastore.experimental.redis import service, system
from trove.guestagent.datastore import manager
from trove.guestagent import guest_log
from trove.guestagent import volume


LOG = logging.getLogger(__name__)


class Manager(manager.Manager):
    """
    This is the Redis manager class. It is dynamically loaded
    based off of the service_type of the trove instance
    """

    GUEST_LOG_DEFS_REDIS_LABEL = 'redis'

    def __init__(self):
        super(Manager, self).__init__('redis')
        self._app = service.RedisApp()

    @property
    def status(self):
        return self._app.status

    @property
    def configuration_manager(self):
        return self._app.configuration_manager

    def get_datastore_log_defs(self):
        logfile = self._app.get_logfile()
        if not logfile:
            return {}
        return {
            self.GUEST_LOG_DEFS_REDIS_LABEL: {
                self.GUEST_LOG_TYPE_LABEL: guest_log.LogType.SYS,
                self.GUEST_LOG_USER_LABEL: system.REDIS_OWNER,
                self.GUEST_LOG_FILE_LABEL: logfile
            }
        }

    def _perform_restore(self, backup_info, context, restore_location, app):
        """Perform a restore on this instance."""
        LOG.info("Restoring database from backup %s.", backup_info['id'])
        try:
            backup.restore(context, backup_info, restore_location)
        except Exception:
            LOG.exception("Error performing restore from backup %s.",
                          backup_info['id'])
            app.status.set_status(rd_instance.ServiceStatuses.FAILED)
            raise
        LOG.info("Restored database successfully.")

    def do_prepare(self, context, packages, databases, memory_mb, users,
                   device_path, mount_point, backup_info,
                   config_contents, root_password, overrides,
                   cluster_config, snapshot):
        """This is called from prepare in the base class."""
        if device_path:
            device = volume.VolumeDevice(device_path)
            # unmount if device is already mounted
            device.unmount_device(device_path)
            device.format()
            device.mount(mount_point)
            operating_system.chown(mount_point, 'redis', 'redis',
                                   as_root=True)
            LOG.debug('Mounted the volume.')
        self._app.install_if_needed(packages)
        LOG.info('Writing redis configuration.')
        if cluster_config:
            config_contents = (config_contents + "\n"
                               + "cluster-enabled yes\n"
                               + "cluster-config-file cluster.conf\n")
        self._app.configuration_manager.save_configuration(config_contents)
        self._app.apply_initial_guestagent_configuration()
        if backup_info:
            persistence_dir = self._app.get_working_dir()
            self._perform_restore(backup_info, context, persistence_dir,
                                  self._app)
        else:
            # If we're not restoring, we have to force a restart of the
            # server manually so that the configuration stuff takes effect
            self._app.restart()
        if snapshot:
            self.attach_replica(context, snapshot, snapshot['config'])

    def pre_upgrade(self, context):
        mount_point = self._app.get_working_dir()
        save_etc_dir = "%s/etc" % mount_point
        home_save = "%s/trove_user" % mount_point

        self._app.status.begin_restart()
        self._app.stop_db()

        operating_system.copy("%s/." % system.REDIS_CONF_DIR, save_etc_dir,
                              preserve=True, as_root=True)

        operating_system.copy("%s/." % os.path.expanduser('~'), home_save,
                              preserve=True, as_root=True)

        self.unmount_volume(context, mount_point=mount_point)

        return {
            'mount_point': mount_point,
            'save_etc_dir': save_etc_dir,
            'home_save': home_save
        }

    def post_upgrade(self, context, upgrade_info):
        self._app.stop_db()

        if 'device' in upgrade_info:
            self.mount_volume(context, mount_point=upgrade_info['mount_point'],
                              device_path=upgrade_info['device'],
                              write_to_fstab=True)
            operating_system.chown(path=upgrade_info['mount_point'],
                                   user=system.REDIS_OWNER,
                                   group=system.REDIS_OWNER,
                                   recursive=True, as_root=True)

        self._restore_home_directory(upgrade_info['home_save'])

        self._restore_directory(upgrade_info['save_etc_dir'],
                                system.REDIS_CONF_DIR)

        self._app = service.RedisApp()
        self._app.start_db()
        self._app.status.end_restart()

    def restart(self, context):
        """
        Restart this redis instance.
        This method is called when the guest agent
        gets a restart message from the taskmanager.
        """
        LOG.debug("Restart called.")
        self._app.restart()

    def start_db_with_conf_changes(self, context, config_contents):
        """
        Start this redis instance with new conf changes.
        """
        LOG.debug("Start DB with conf changes called.")
        self._app.start_db_with_conf_changes(config_contents)

    def stop_db(self, context, do_not_start_on_reboot=False):
        """
        Stop this redis instance.
        This method is called when the guest agent
        gets a stop message from the taskmanager.
        """
        LOG.debug("Stop DB called.")
        self._app.stop_db(do_not_start_on_reboot=do_not_start_on_reboot)

    def create_backup(self, context, backup_info):
        """Create a backup of the database."""
        LOG.debug("Creating backup.")
        with EndNotification(context):
            backup.backup(context, backup_info)

    def update_overrides(self, context, overrides, remove=False):
        LOG.debug("Updating overrides.")
        if remove:
            self._app.remove_overrides()
        else:
            self._app.update_overrides(context, overrides, remove)

    def apply_overrides(self, context, overrides):
        LOG.debug("Applying overrides.")
        self._app.apply_overrides(self._app.admin, overrides)

    def backup_required_for_replication(self, context):
        return self.replication.backup_required_for_replication()

    def get_replication_snapshot(self, context, snapshot_info,
                                 replica_source_config=None):
        LOG.debug("Getting replication snapshot.")
        self.replication.enable_as_master(self._app, replica_source_config)

        snapshot_id, log_position = self.replication.snapshot_for_replication(
            context, self._app, None, snapshot_info)

        volume_stats = self.get_filesystem_stats(context, None)

        replication_snapshot = {
            'dataset': {
                'datastore_manager': self.manager,
                'dataset_size': volume_stats.get('used', 0.0),
                'volume_size': volume_stats.get('total', 0.0),
                'snapshot_id': snapshot_id
            },
            'replication_strategy': self.replication_strategy,
            'master': self.replication.get_master_ref(self._app,
                                                      snapshot_info),
            'log_position': log_position
        }

        return replication_snapshot

    def enable_as_master(self, context, replica_source_config):
        LOG.debug("Calling enable_as_master.")
        self.replication.enable_as_master(self._app, replica_source_config)

    def detach_replica(self, context, for_failover=False):
        LOG.debug("Detaching replica.")
        replica_info = self.replication.detach_slave(self._app, for_failover)
        return replica_info

    def get_replica_context(self, context):
        LOG.debug("Getting replica context.")
        replica_info = self.replication.get_replica_context(self._app)
        return replica_info

    def _validate_slave_for_replication(self, context, replica_info):
        if replica_info['replication_strategy'] != self.replication_strategy:
            raise exception.IncompatibleReplicationStrategy(
                replica_info.update({
                    'guest_strategy': self.replication_strategy
                }))

    def attach_replica(self, context, replica_info, slave_config):
        LOG.debug("Attaching replica.")
        try:
            if 'replication_strategy' in replica_info:
                self._validate_slave_for_replication(context, replica_info)
            self.replication.enable_as_slave(self._app, replica_info,
                                             slave_config)
        except Exception:
            LOG.exception("Error enabling replication.")
            raise

    def make_read_only(self, context, read_only):
        LOG.debug("Executing make_read_only(%s)", read_only)
        self._app.make_read_only(read_only)

    def _get_repl_info(self):
        return self._app.admin.get_info('replication')

    def _get_master_host(self):
        slave_info = self._get_repl_info()
        return slave_info and slave_info['master_host'] or None

    def _get_repl_offset(self):
        repl_info = self._get_repl_info()
        LOG.debug("Got repl info: %s", repl_info)
        offset_key = '%s_repl_offset' % repl_info['role']
        offset = repl_info[offset_key]
        LOG.debug("Found offset %(offset)s for key %(key)s.",
                  {'offset': offset, 'key': offset_key})
        return int(offset)

    def get_last_txn(self, context):
        master_host = self._get_master_host()
        repl_offset = self._get_repl_offset()
        return master_host, repl_offset

    def get_latest_txn_id(self, context):
        LOG.info("Retrieving latest repl offset.")
        return self._get_repl_offset()

    def wait_for_txn(self, context, txn):
        LOG.info("Waiting on repl offset '%s'.", txn)

        def _wait_for_txn():
            current_offset = self._get_repl_offset()
            LOG.debug("Current offset: %s.", current_offset)
            return current_offset >= txn

        try:
            utils.poll_until(_wait_for_txn, time_out=120)
        except exception.PollTimeOut:
            raise RuntimeError(_("Timeout occurred waiting for Redis repl "
                                 "offset to change to '%s'.") % txn)

    def cleanup_source_on_replica_detach(self, context, replica_info):
        LOG.debug("Cleaning up the source on the detach of a replica.")
        self.replication.cleanup_source_on_replica_detach(self._app,
                                                          replica_info)

    def demote_replication_master(self, context):
        LOG.debug("Demoting replica source.")
        self.replication.demote_master(self._app)

    def cluster_meet(self, context, ip, port):
        LOG.debug("Executing cluster_meet to join node to cluster.")
        self._app.cluster_meet(ip, port)

    def get_node_ip(self, context):
        LOG.debug("Retrieving cluster node ip address.")
        return self._app.get_node_ip()

    def get_node_id_for_removal(self, context):
        LOG.debug("Validating removal of node from cluster.")
        return self._app.get_node_id_for_removal()

    def remove_nodes(self, context, node_ids):
        LOG.debug("Removing nodes from cluster.")
        self._app.remove_nodes(node_ids)

    def cluster_addslots(self, context, first_slot, last_slot):
        LOG.debug("Executing cluster_addslots to assign hash slots %s-%s.",
                  first_slot, last_slot)
        self._app.cluster_addslots(first_slot, last_slot)

    def enable_root(self, context):
        LOG.debug("Enabling authentication.")
        return self._app.enable_root()

    def enable_root_with_password(self, context, root_password=None):
        LOG.debug("Enabling authentication with password.")
        return self._app.enable_root(root_password)

    def disable_root(self, context):
        LOG.debug("Disabling authentication.")
        return self._app.disable_root()

    def get_root_password(self, context):
        LOG.debug("Getting auth password.")
        return self._app.get_auth_password()



# Copyright 2019 Catalyst Cloud Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import netaddr
from oslo_log import log as logging

from trove.common import cfg
from trove.common import clients
from trove.common import exception

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
MGMT_NETWORKS = None
MGMT_CIDRS = None


def get_management_networks(context):
    """Cache the management network names.

    When CONF.management_networks is changed, the Trove service needs to
    restart so the global cache will be refreshed.
    """
    global MGMT_NETWORKS

    if MGMT_NETWORKS is not None:
        return MGMT_NETWORKS

    MGMT_NETWORKS = []
    if len(CONF.management_networks) > 0:
        neutron_client = clients.create_neutron_client(context)

        for net_id in CONF.management_networks:
            MGMT_NETWORKS.append(
                neutron_client.show_network(net_id)['network']['name']
            )

    return MGMT_NETWORKS


def reset_management_networks():
    """This method is only for testing purpose."""
    global MGMT_NETWORKS

    MGMT_NETWORKS = None


def create_port(client, name, description, network_id, security_groups,
                is_public=False):
    port_body = {
        "port": {
            "name": name,
            "description": description,
            "network_id": network_id,
            "security_groups": security_groups
        }
    }
    port = client.create_port(body=port_body)
    port_id = port['port']['id']

    if is_public:
        public_network_id = get_public_network(client)
        if not public_network_id:
            raise exception.PublicNetworkNotFound()

        fip_body = {
            "floatingip": {
                'floating_network_id': public_network_id,
                'port_id': port_id,
            }
        }
        client.create_floatingip(fip_body)

    return port_id


def delete_port(client, id):
    ret = client.list_floatingips(port_id=id)
    if len(ret['floatingips']) > 0:
        for fip in ret['floatingips']:
            try:
                client.delete_floatingip(fip['id'])
            except Exception as e:
                LOG.error(
                    'Failed to delete floating IP for port %s, error: %s',
                    id, str(e)
                )

    client.delete_port(id)


def get_public_network(client):
    """Get public network ID.

    If not given in the config file, try to query all the public networks and
    use the first one in the list.
    """
    if CONF.network.public_network_id:
        return CONF.network.public_network_id

    kwargs = {'router:external': True}
    ret = client.list_networks(**kwargs)

    if len(ret.get('networks', [])) == 0:
        return None

    return ret['networks'][0].get('id')


def create_security_group(client, name, instance_id):
    body = {
        'security_group': {
            'name': name,
            'description': 'Security group for trove instance %s' % instance_id
        }
    }
    ret = client.create_security_group(body=body)
    return ret['security_group']['id']


def create_security_group_rule(client, sg_id, protocol, ports, remote_ips):
    for remote_ip in remote_ips:
        ip = netaddr.IPNetwork(remote_ip)
        ethertype = 'IPv4' if ip.version == 4 else 'IPv6'

        for port_or_range in set(ports):
            from_, to_ = port_or_range[0], port_or_range[-1]

            body = {
                "security_group_rule": {
                    "direction": "ingress",
                    "ethertype": ethertype,
                    "protocol": protocol,
                    "security_group_id": sg_id,
                    "port_range_min": int(from_),
                    "port_range_max": int(to_),
                    "remote_ip_prefix": remote_ip
                }
            }

            client.create_security_group_rule(body)


def get_subnet_cidrs(client, network_id):
    cidrs = []

    subnets = client.list_subnets(network_id=network_id)['subnets']
    for subnet in subnets:
        cidrs.append(subnet.get('cidr'))

    return cidrs


def get_mamangement_subnet_cidrs(client):
    """Cache the management subnet CIDRS."""
    global MGMT_CIDRS

    if MGMT_CIDRS is not None:
        return MGMT_CIDRS

    MGMT_CIDRS = []
    if len(CONF.management_networks) > 0:
        MGMT_CIDRS = get_subnet_cidrs(client, CONF.management_networks[0])

    return MGMT_CIDRS



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import getpass
import os

from mock import ANY
from mock import DEFAULT
from mock import MagicMock
from mock import Mock
from mock import patch
from oslo_utils import encodeutils
from proboscis.asserts import assert_equal
from proboscis.asserts import assert_is_none
from proboscis.asserts import assert_true

from trove.common.context import TroveContext
from trove.common import exception
from trove.guestagent.common import operating_system
from trove.guestagent.datastore import manager
from trove.guestagent import guest_log
from trove.guestagent.module import module_manager
from trove import rpc
from trove.tests.unittests import trove_testtools


class MockManager(manager.Manager):

    def __init__(self):
        super(MockManager, self).__init__('mysql')
        self._app = MagicMock()
        self._status = MagicMock()
        self._configuration_manager = MagicMock()

    @property
    def app(self):
        return self._app

    @property
    def status(self):
        return self._status

    @property
    def configuration_manager(self):
        return self._configuration_manager

    def prepare(self, *args):
        args[0].notification = MagicMock()
        with patch.object(rpc, 'get_client'):
            return super(MockManager, self).prepare(*args)


class ManagerTest(trove_testtools.TestCase):

    def setUp(self):
        super(ManagerTest, self).setUp()

        self.chmod_patch = patch.object(operating_system, 'chmod')
        self.chmod_mock = self.chmod_patch.start()
        self.addCleanup(self.chmod_patch.stop)

        self.manager = MockManager()
        self.context = TroveContext()

        self.log_name_sys = 'guest'
        self.log_name_user = 'general'
        self.prefix = 'log_prefix'
        self.container = 'log_container'
        self.size = 1024
        self.published = 128
        self.guest_log_user = guest_log.GuestLog(
            self.context, self.log_name_user, guest_log.LogType.USER, None,
            '/tmp/gen.log', True)
        self.guest_log_sys = guest_log.GuestLog(
            self.context, self.log_name_sys, guest_log.LogType.SYS, None,
            '/tmp/guest.log', True)
        for gl in [self.guest_log_user, self.guest_log_sys]:
            gl._container_name = self.container
            gl._refresh_details = MagicMock()
            gl._log_rotated = MagicMock(return_value=False)
            gl._publish_to_container = MagicMock()
            gl._delete_log_components = MagicMock()
            gl._object_prefix = MagicMock(return_value=self.prefix)
            gl._size = self.size
            gl._published_size = self.published
        self.manager._guest_log_cache = {
            self.log_name_user: self.guest_log_user,
            self.log_name_sys: self.guest_log_sys}
        self.expected_details_user = {
            'status': 'Disabled',
            'prefix': self.prefix,
            'container': self.container,
            'name': self.log_name_user,
            'published': self.published,
            'metafile': self.prefix + '_metafile',
            'type': 'USER',
            'pending': self.size - self.published}
        self.expected_details_sys = dict(self.expected_details_user)
        self.expected_details_sys['type'] = 'SYS'
        self.expected_details_sys['status'] = 'Enabled'
        self.expected_details_sys['name'] = self.log_name_sys
        self.expected_module_details = {
            'name': 'mymod',
            'type': 'ping',
            'contents': 'e262cfe36134'
        }
        self.manager.module_manager = Mock()

    def tearDown(self):
        super(ManagerTest, self).tearDown()

    def test_update_status(self):
        self.manager._status.is_installed = True
        self.manager._status._is_restarting = False
        self.manager.update_status(self.context)
        self.assertTrue(self.manager.status.set_status.called)

    def test_guest_log_list(self):
        log_list = self.manager.guest_log_list(self.context)
        expected = [self.expected_details_sys, self.expected_details_user]
        assert_equal(self._flatten_list_of_dicts(expected),
                     self._flatten_list_of_dicts(log_list),
                     "Wrong list: %s (Expected: %s)" % (
                         self._flatten_list_of_dicts(log_list),
                         self._flatten_list_of_dicts(expected)))

    def _flatten_list_of_dicts(self, lod):
        value = sorted("".join("%s%s" % (k, d[k]) for k in sorted(d.keys()))
                       for d in lod)
        return "".join(sorted(value))

    def test_guest_log_action_enable_disable(self):
        self.assertRaisesRegex(exception.BadRequest,
                               "Cannot enable and disable",
                               self.manager.guest_log_action,
                               self.context,
                               self.log_name_sys,
                               True, True, False, False)

    def test_guest_log_action_enable_sys(self):
        self.assertRaisesRegex(exception.BadRequest,
                               "Cannot enable a SYSTEM log",
                               self.manager.guest_log_action,
                               self.context,
                               self.log_name_sys,
                               True, False, False, False)

    def test_guest_log_action_disable_sys(self):
        self.assertRaisesRegex(exception.BadRequest,
                               "Cannot disable a SYSTEM log",
                               self.manager.guest_log_action,
                               self.context,
                               self.log_name_sys,
                               False, True, False, False)

    def test_guest_log_action_publish_sys(self):
        with patch.object(os.path, 'isfile', return_value=True):
            log_details = self.manager.guest_log_action(self.context,
                                                        self.log_name_sys,
                                                        False, False,
                                                        True, False)
            assert_equal(log_details, self.expected_details_sys,
                         "Wrong details: %s (expected %s)" %
                         (log_details, self.expected_details_sys))
            assert_equal(
                1, self.guest_log_sys._publish_to_container.call_count)

    def test_guest_log_action_discard_sys(self):
        log_details = self.manager.guest_log_action(self.context,
                                                    self.log_name_sys,
                                                    False, False,
                                                    False, True)
        assert_equal(log_details, self.expected_details_sys,
                     "Wrong details: %s (expected %s)" %
                     (log_details, self.expected_details_sys))
        assert_equal(
            1, self.guest_log_sys._delete_log_components.call_count)

    def test_guest_log_action_enable_user(self):
        with patch.object(manager.Manager, 'guest_log_enable',
                          return_value=False) as mock_enable:
            log_details = self.manager.guest_log_action(self.context,
                                                        self.log_name_user,
                                                        True, False,
                                                        False, False)
            assert_equal(log_details, self.expected_details_user,
                         "Wrong details: %s (expected %s)" %
                         (log_details, self.expected_details_user))
            assert_equal(1, mock_enable.call_count)

    def test_guest_log_action_disable_user(self):
        with patch.object(manager.Manager, 'guest_log_enable',
                          return_value=False) as mock_enable:
            self.guest_log_user._enabled = True
            log_details = self.manager.guest_log_action(self.context,
                                                        self.log_name_user,
                                                        False, True,
                                                        False, False)
            assert_equal(log_details, self.expected_details_user,
                         "Wrong details: %s (expected %s)" %
                         (log_details, self.expected_details_user))
            assert_equal(1, mock_enable.call_count)

    def test_guest_log_action_publish_user(self):
        with patch.object(manager.Manager, 'guest_log_enable',
                          return_value=False) as mock_enable:
            with patch.object(os.path, 'isfile', return_value=True):
                log_details = self.manager.guest_log_action(self.context,
                                                            self.log_name_user,
                                                            False, False,
                                                            True, False)
                assert_equal(log_details, self.expected_details_user,
                             "Wrong details: %s (expected %s)" %
                             (log_details, self.expected_details_user))
                assert_equal(1, mock_enable.call_count)

    def test_guest_log_action_discard_user(self):
        log_details = self.manager.guest_log_action(self.context,
                                                    self.log_name_user,
                                                    False, False,
                                                    False, True)
        assert_equal(log_details, self.expected_details_user,
                     "Wrong details: %s (expected %s)" %
                     (log_details, self.expected_details_user))
        assert_equal(1, self.guest_log_user._delete_log_components.call_count)

    def test_set_guest_log_status_disabled(self):
        data = [
            {'orig': guest_log.LogStatus.Enabled,
             'new': guest_log.LogStatus.Disabled,
             'expect': guest_log.LogStatus.Disabled},
            {'orig': guest_log.LogStatus.Restart_Required,
             'new': guest_log.LogStatus.Enabled,
             'expect': guest_log.LogStatus.Restart_Required},
            {'orig': guest_log.LogStatus.Restart_Required,
             'new': guest_log.LogStatus.Restart_Completed,
             'expect': guest_log.LogStatus.Restart_Completed},
            {'orig': guest_log.LogStatus.Published,
             'new': guest_log.LogStatus.Partial,
             'expect': guest_log.LogStatus.Partial},
        ]
        for datum in data:
            self.assert_guest_log_status(datum['orig'],
                                         datum['new'],
                                         datum['expect'])

    def assert_guest_log_status(self, original_status, new_status,
                                expected_final_status):
        gl_cache = self.manager.get_guest_log_cache()
        gl_cache[self.log_name_sys]._status = original_status
        self.manager.set_guest_log_status(new_status, self.log_name_sys)
        assert_equal(gl_cache[self.log_name_sys].status, expected_final_status,
                     "Unexpected status for '%s': %s' (Expected %s)" %
                     (self.log_name_sys, gl_cache[self.log_name_sys].status,
                      expected_final_status))

    def test_build_log_file_name(self):
        current_owner = getpass.getuser()
        with patch.multiple(operating_system,
                            exists=MagicMock(return_value=False),
                            write_file=DEFAULT,
                            create_directory=DEFAULT,
                            chown=DEFAULT,
                            chmod=DEFAULT) as os_mocks:
            log_file = self.manager.build_log_file_name(self.log_name_sys,
                                                        current_owner)
            expected_filename = '%s/%s/%s-%s.log' % (
                self.manager.GUEST_LOG_BASE_DIR,
                self.manager.GUEST_LOG_DATASTORE_DIRNAME,
                self.manager.manager, self.log_name_sys)
            expected_call_counts = {'exists': 1,
                                    'write_file': 1,
                                    'create_directory': 2,
                                    'chown': 1,
                                    'chmod': 1}
            self.assert_build_log_file_name(expected_filename, log_file,
                                            os_mocks, expected_call_counts)

    def assert_build_log_file_name(self, expected_filename, filename,
                                   mocks, call_counts):
        assert_equal(expected_filename, filename,
                     "Unexpected filename: %s (expected %s)" %
                     (filename, expected_filename))
        for key in mocks.keys():
            assert_true(
                mocks[key].call_count == call_counts[key],
                "%s called %d time(s)" % (key, mocks[key].call_count))

    def test_build_log_file_name_with_dir(self):
        current_owner = getpass.getuser()
        log_dir = '/tmp'
        with patch.multiple(operating_system,
                            exists=MagicMock(return_value=False),
                            write_file=DEFAULT,
                            create_directory=DEFAULT,
                            chown=DEFAULT,
                            chmod=DEFAULT) as os_mocks:
            log_file = self.manager.build_log_file_name(self.log_name_sys,
                                                        current_owner,
                                                        datastore_dir=log_dir)
            expected_filename = '%s/%s-%s.log' % (
                log_dir,
                self.manager.manager, self.log_name_sys)
            expected_call_counts = {'exists': 1,
                                    'write_file': 1,
                                    'create_directory': 1,
                                    'chown': 1,
                                    'chmod': 1}
            self.assert_build_log_file_name(expected_filename, log_file,
                                            os_mocks, expected_call_counts)

    def test_validate_log_file(self):
        file_name = '/tmp/non-existent-file'
        current_owner = getpass.getuser()
        with patch.multiple(operating_system,
                            exists=MagicMock(return_value=False),
                            write_file=DEFAULT,
                            chown=DEFAULT,
                            chmod=DEFAULT) as os_mocks:
            log_file = self.manager.validate_log_file(file_name, current_owner)
            assert_equal(file_name, log_file, "Unexpected filename")
            for key in os_mocks.keys():
                assert_true(os_mocks[key].call_count == 1,
                            "%s not called" % key)

    def test_prepare_single(self):
        self.run_prepare_test(cluster_config=None)

    def test_prepare_single_no_users(self):
        self.run_prepare_test(cluster_config=None, users=None)

    def test_prepare_single_no_databases(self):
        self.run_prepare_test(cluster_config=None, databases=None)

    def test_prepare_single_no_root_password(self):
        self.run_prepare_test(cluster_config=None, root_password=None)

    def test_prepare_cluster(self):
        self.run_prepare_test()

    def run_prepare_test(self, packages=Mock(), databases=Mock(),
                         memory_mb=Mock(), users=Mock(),
                         device_path=Mock(), mount_point=Mock(),
                         backup_info=Mock(), config_contents=Mock(),
                         root_password=Mock(), overrides=Mock(),
                         cluster_config=Mock(), snapshot=Mock()):
        self._assert_prepare(self.context, packages, databases, memory_mb,
                             users, device_path, mount_point, backup_info,
                             config_contents, root_password, overrides,
                             cluster_config, snapshot)

    def _assert_prepare(self, context, packages, databases, memory_mb, users,
                        device_path, mount_point, backup_info, config_contents,
                        root_password, overrides, cluster_config, snapshot):

        is_error_expected = False
        is_post_process_expected = cluster_config is not None

        with patch.multiple(self.manager,
                            do_prepare=DEFAULT, post_prepare=DEFAULT,
                            apply_overrides_on_prepare=DEFAULT,
                            enable_root_on_prepare=DEFAULT,
                            create_database=DEFAULT, create_user=DEFAULT):
            self.manager.prepare(
                context, packages, databases, memory_mb, users,
                device_path, mount_point, backup_info, config_contents,
                root_password, overrides, cluster_config, snapshot)

            self.manager.status.begin_install.assert_called_once_with()
            self.manager.do_prepare.assert_called_once_with(
                context,
                packages,
                databases,
                memory_mb,
                users,
                device_path,
                mount_point,
                backup_info,
                config_contents,
                root_password,
                overrides,
                cluster_config,
                snapshot)
            self.manager.apply_overrides_on_prepare.assert_called_once_with(
                context,
                overrides)
            self.manager.status.end_install(
                error_occurred=is_error_expected,
                post_processing=is_post_process_expected)
            self.manager.post_prepare.assert_called_once_with(
                context,
                packages,
                databases,
                memory_mb,
                users,
                device_path,
                mount_point,
                backup_info,
                config_contents,
                root_password,
                overrides,
                cluster_config,
                snapshot)

            if not is_post_process_expected:
                if databases:
                    self.manager.create_database.assert_called_once_with(
                        context,
                        databases)
                else:
                    self.assertEqual(
                        0, self.manager.create_database.call_count)
                if users:
                    self.manager.create_user.assert_called_once_with(
                        context,
                        users)
                else:
                    self.assertEqual(0, self.manager.create_user.call_count)
                if not backup_info and root_password:
                    (self.manager.enable_root_on_prepare.
                     assert_called_once_with(context, root_password))
                else:
                    self.assertEqual(
                        0, self.manager.enable_root_on_prepare.call_count)
            else:
                self.assertEqual(0, self.manager.create_database.call_count)
                self.assertEqual(0, self.manager.create_user.call_count)
                self.assertEqual(
                    0, self.manager.enable_root_on_prepare.call_count)

    def test_apply_overrides_on_prepare(self):
        overrides = Mock()
        with patch.multiple(self.manager,
                            update_overrides=DEFAULT, restart=DEFAULT):
            self.manager.apply_overrides_on_prepare(self.context, overrides)

            self.manager.update_overrides.assert_called_once_with(
                self.context, overrides)
            self.manager.restart.assert_called_once_with(self.context)

    @patch('trove.guestagent.datastore.manager.LOG')
    def test_apply_overrides_on_prepare_failure(self, mock_logging):
        packages = Mock()
        databases = Mock()
        memory_mb = Mock()
        users = Mock()
        device_path = Mock()
        mount_point = Mock()
        backup_info = Mock()
        config_contents = Mock()
        root_password = Mock()
        overrides = Mock()
        cluster_config = Mock()
        snapshot = Mock()

        expected_failure = Exception("Error in 'apply_overrides_on_prepare'.")

        with patch.multiple(
                self.manager, do_prepare=DEFAULT,
                apply_overrides_on_prepare=MagicMock(
                    side_effect=expected_failure
                )):
            expected_msg = encodeutils.exception_to_unicode(expected_failure)
            self.assertRaisesRegex(
                Exception, expected_msg,
                self.manager.prepare,
                self.context, packages, databases, memory_mb, users,
                device_path, mount_point, backup_info, config_contents,
                root_password, overrides, cluster_config, snapshot)

            self.manager.status.begin_install.assert_called_once_with()
            self.manager.status.end_install(
                error_occurred=True,
                post_processing=ANY)

    @patch.object(operating_system, 'copy')
    @patch.object(operating_system, 'chown')
    def test_restore_directory_with_owner(self, chown_mock, copy_mock):
        restore_dir = '/restore_directory'
        restore_files = '/restore_directory/.'
        target_dir = '/target_directory'
        owner = 'owner'
        self.manager._restore_directory(restore_dir, target_dir, owner)
        copy_mock.assert_called_once_with(restore_files, target_dir,
                                          preserve=True, as_root=True)
        chown_mock.assert_called_once_with(path=target_dir, user=owner,
                                           group=owner, recursive=True,
                                           as_root=True)

    @patch.object(operating_system, 'copy')
    @patch.object(operating_system, 'chown')
    def test_restore_directory_without_owner(self, chown_mock, copy_mock):
        restore_dir = '/restore_directory'
        restore_files = '/restore_directory/.'
        target_dir = '/target_directory'
        self.manager._restore_directory(restore_dir, target_dir)
        copy_mock.assert_called_once_with(restore_files, target_dir,
                                          preserve=True, as_root=True)
        chown_mock.assert_not_called()

    @patch.object(manager.Manager, '_restore_directory')
    @patch.object(operating_system, 'get_current_user', return_value='trove')
    def test_restore_home_directory(self, os_mock, restore_mock):
        saved_home_dir = '/old_home'
        with patch.object(os.path, 'expanduser', return_value='/home/trove'):
            self.manager._restore_home_directory(saved_home_dir)
            os_mock.assert_any_call()
            restore_mock.assert_called_once_with(restore_dir=saved_home_dir,
                                                 target_dir='/home/trove',
                                                 owner='trove')

    def test_module_list(self):
        with patch.object(module_manager.ModuleManager, 'read_module_results',
                          return_value=[
                              self.expected_module_details]) as mock_rmr:
            module_list = self.manager.module_list(self.context)
            expected = [self.expected_module_details]
            assert_equal(self._flatten_list_of_dicts(expected),
                         self._flatten_list_of_dicts(module_list),
                         "Wrong list: %s (Expected: %s)" % (
                             self._flatten_list_of_dicts(module_list),
                             self._flatten_list_of_dicts(expected)))
            assert_equal(1, mock_rmr.call_count)

    def test_module_apply(self):
        with patch.object(
                module_manager.ModuleManager, 'apply_module',
                return_value=[self.expected_module_details]) as mock_am:
            module_details = self.manager.module_apply(
                self.context,
                [{'module': self.expected_module_details}])
            assert_equal([[self.expected_module_details]], module_details)
            assert_equal(1, mock_am.call_count)

    def test_module_remove(self):
        with patch.object(
                module_manager.ModuleManager, 'remove_module',
                return_value=[self.expected_module_details]) as mock_rm:
            module_details = self.manager.module_remove(
                self.context,
                {'module': self.expected_module_details})
            assert_is_none(module_details)
            assert_equal(1, mock_rm.call_count)



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import re

from eventlet.green import subprocess
from oslo_log import log as logging

from trove.common import cfg
from trove.common import stream_codecs
from trove.guestagent.common import operating_system
from trove.guestagent.common.operating_system import FileMode
from trove.guestagent.datastore.experimental.postgresql.service import PgSqlApp
from trove.guestagent.strategies.restore import base

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
WAL_ARCHIVE_DIR = CONF.postgresql.wal_archive_location


class PgDump(base.RestoreRunner):
    """Implementation of Restore Strategy for pg_dump."""
    __strategy_name__ = 'pg_dump'
    base_restore_cmd = 'psql -U os_admin'

    IGNORED_ERROR_PATTERNS = [
        re.compile(r'ERROR:\s*role "postgres" already exists'),
    ]

    def restore(self):
        """We are overriding the base class behavior
        to perform custom error handling.
        """
        self.pre_restore()
        content_length = self._execute_postgres_restore()
        self.post_restore()
        return content_length

    def _execute_postgres_restore(self):
        # Postgresql outputs few benign messages into the stderr stream
        # during a normal restore procedure.
        # We need to watch for those and avoid raising
        # an exception in response.
        # Message 'ERROR:  role "postgres" already exists'
        # is expected and does not pose any problems to the restore operation.

        stream = self.storage.load(self.location, self.checksum)
        process = subprocess.Popen(self.restore_cmd, shell=True,
                                   stdin=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        content_length = 0
        for chunk in stream:
            process.stdin.write(chunk)
            content_length += len(chunk)
        process.stdin.close()
        self._handle_errors(process)
        LOG.info("Restored %s bytes from stream.", content_length)

        return content_length

    def _handle_errors(self, process):
        # Handle messages in the error stream of a given process.
        # Raise an exception if the stream is not empty and
        # does not match the expected message sequence.

        try:
            err = process.stderr.read()
            # Empty error stream is always accepted as valid
            # for future compatibility.
            if err:
                for message in err.splitlines(False):
                    if not any(regex.match(message)
                               for regex in self.IGNORED_ERROR_PATTERNS):
                        raise Exception(message)
        except OSError:
            pass


class PgBaseBackup(base.RestoreRunner):
    """Implementation of Restore Strategy for pg_basebackup."""
    __strategy_name__ = 'pg_basebackup'
    location = ""
    base_restore_cmd = ''

    IGNORED_ERROR_PATTERNS = [
        re.compile(r'ERROR:\s*role "postgres" already exists'),
    ]

    def __init__(self, *args, **kwargs):
        self._app = None
        self.base_restore_cmd = 'sudo -u %s tar xCf %s - ' % (
            self.app.pgsql_owner, self.app.pgsql_data_dir
        )

        super(PgBaseBackup, self).__init__(*args, **kwargs)

    @property
    def app(self):
        if self._app is None:
            self._app = self._build_app()
        return self._app

    def _build_app(self):
        return PgSqlApp()

    def pre_restore(self):
        self.app.stop_db()
        LOG.info("Preparing WAL archive dir")
        self.app.recreate_wal_archive_dir()
        datadir = self.app.pgsql_data_dir
        operating_system.remove(datadir, force=True, recursive=True,
                                as_root=True)
        operating_system.create_directory(datadir, user=self.app.pgsql_owner,
                                          group=self.app.pgsql_owner,
                                          force=True, as_root=True)

    def post_restore(self):
        operating_system.chmod(self.app.pgsql_data_dir,
                               FileMode.SET_USR_RWX(),
                               as_root=True, recursive=True, force=True)

    def write_recovery_file(self, restore=False):
        metadata = self.storage.load_metadata(self.location, self.checksum)
        recovery_conf = ""
        recovery_conf += "recovery_target_name = '%s' \n" % metadata['label']
        recovery_conf += "recovery_target_timeline = '%s' \n" % 1

        if restore:
            recovery_conf += ("restore_command = '" +
                              self.pgsql_restore_cmd + "'\n")

        recovery_file = os.path.join(self.app.pgsql_data_dir, 'recovery.conf')
        operating_system.write_file(recovery_file, recovery_conf,
                                    codec=stream_codecs.IdentityCodec(),
                                    as_root=True)
        operating_system.chown(recovery_file, user=self.app.pgsql_owner,
                               group=self.app.pgsql_owner, as_root=True)


class PgBaseBackupIncremental(PgBaseBackup):

    def __init__(self, *args, **kwargs):
        super(PgBaseBackupIncremental, self).__init__(*args, **kwargs)
        self.content_length = 0
        self.incr_restore_cmd = 'sudo -u %s tar -xf - -C %s ' % (
                                self.app.pgsql_owner, WAL_ARCHIVE_DIR
        )
        self.pgsql_restore_cmd = "cp " + WAL_ARCHIVE_DIR + '/%f "%p"'

    def pre_restore(self):
        self.app.stop_db()

    def post_restore(self):
        self.write_recovery_file(restore=True)

    def _incremental_restore_cmd(self, incr=False):
        args = {'restore_location': self.restore_location}
        cmd = self.base_restore_cmd
        if incr:
            cmd = self.incr_restore_cmd
        return self.decrypt_cmd + self.unzip_cmd + (cmd % args)

    def _incremental_restore(self, location, checksum):

        metadata = self.storage.load_metadata(location, checksum)
        if 'parent_location' in metadata:
            LOG.info("Found parent at %s", metadata['parent_location'])
            parent_location = metadata['parent_location']
            parent_checksum = metadata['parent_checksum']
            self._incremental_restore(parent_location, parent_checksum)
            cmd = self._incremental_restore_cmd(incr=True)
            self.content_length += self._unpack(location, checksum, cmd)

        else:
            # For the parent base backup, revert to the default restore cmd
            LOG.info("Recursed back to full backup.")

            super(PgBaseBackupIncremental, self).pre_restore()
            cmd = self._incremental_restore_cmd(incr=False)
            self.content_length += self._unpack(location, checksum, cmd)

            operating_system.chmod(self.app.pgsql_data_dir,
                                   FileMode.SET_USR_RWX(),
                                   as_root=True, recursive=True, force=True)

    def _run_restore(self):
        self._incremental_restore(self.location, self.checksum)
        # content-length restored
        return self.content_length



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import json
import time

from oslo_log import log as logging
from proboscis import asserts
from proboscis.asserts import assert_equal
from proboscis.asserts import assert_true
from proboscis.asserts import Check
from proboscis import before_class
from proboscis import SkipTest
from proboscis import test
from proboscis import TestProgram
import six
from troveclient.compat import client as trove_client
from troveclient.compat import Dbaas
from troveclient.compat import TroveHTTPClient

from trove.tests.config import CONFIG
from trove.tests.examples.client import JsonClient
from trove.tests.examples.client import SnippetWriter

# troveclient.compat.client._logger was changed to LOG in 2.11.0
if hasattr(trove_client, '_logger'):
    trove_client._logger.setLevel(logging.CRITICAL)
elif hasattr(trove_client, 'LOG'):
    trove_client.LOG.setLevel(logging.CRITICAL)

FAKE_INFO = {'m': 30, 's': 0, 'uuid': 'abcdef00-aaaa-aaaa-aaaa-bbbbbbbbbbbb'}
EXAMPLE_BACKUP_ID = "a9832168-7541-4536-b8d9-a8a9b79cf1b4"
EXAMPLE_BACKUP_INCREMENTAL_ID = "2e351a71-dd28-4bcb-a7d6-d36a5b487173"
EXAMPLE_CONFIG_ID = "43a6ea86-e959-4735-9e46-a6a5d4a2d80f"
EXAMPLE_INSTANCE_ID = "44b277eb-39be-4921-be31-3d61b43651d7"
EXAMPLE_INSTANCE_ID_2 = "d5a9db64-7ef7-41c5-8e1e-4013166874bc"
EXAMPLE_CONFIG_SERVER_ID = "271898715"


def get_now():
    from datetime import datetime
    return datetime(2014, 10, 30, hour=12, minute=FAKE_INFO['m'],
                    second=FAKE_INFO['s'])


def get_uuid():
    return FAKE_INFO['uuid']


def set_fake_stuff(uuid=None, minute=None, unique_id=None):
    if uuid:
        FAKE_INFO['uuid'] = uuid
    if minute:
        FAKE_INFO['minute'] = minute
    if unique_id:
        from trove.common.template import SingleInstanceConfigTemplate

        def fake_calc_id(self):
            return unique_id

        SingleInstanceConfigTemplate._calculate_unique_id = fake_calc_id


def monkey_patch_uuid_and_date():
    import uuid
    uuid.uuid4 = get_uuid
    from trove.common import timeutils
    from trove.common import utils
    timeutils.utcnow = get_now
    utils.generate_uuid = get_uuid


@test
def load_config_file():
    global conf
    if CONFIG.get("examples", None) is None:
        fail("Missing 'examples' config in test config.")
    conf = CONFIG.examples
    global normal_user
    normal_user = CONFIG.users.find_user_by_name(conf['normal_user_name'])
    global admin_user
    admin_user = CONFIG.users.find_user_by_name(conf['admin_user_name'])


def create_client_args(user):

    auth_strategy = None

    kwargs = {
        'service_type': 'trove',
        'insecure': CONFIG.values['trove_client_insecure'],
    }

    def set_optional(kwargs_name, test_conf_name):
        value = CONFIG.values.get(test_conf_name, None)
        if value is not None:
            kwargs[kwargs_name] = value

    service_url = CONFIG.get('override_trove_api_url', None)
    if user.requirements.is_admin:
        service_url = CONFIG.get('override_admin_trove_api_url',
                                 service_url)
    if service_url:
        kwargs['service_url'] = service_url

    auth_strategy = None
    if user.requirements.is_admin:
        auth_strategy = CONFIG.get('admin_auth_strategy',
                                   CONFIG.auth_strategy)
    else:
        auth_strategy = CONFIG.auth_strategy
    set_optional('region_name', 'trove_client_region_name')
    if CONFIG.values.get('override_trove_api_url_append_tenant',
                         False):
        kwargs['service_url'] += "/" + user.tenant

    if auth_strategy == 'fake':
        from troveclient.compat import auth

        class FakeAuth(auth.Authenticator):

            def authenticate(self):
                class FakeCatalog(object):
                    def __init__(self, auth):
                        self.auth = auth

                    def get_public_url(self):
                        return "%s/%s" % (CONFIG.dbaas_url,
                                          self.auth.tenant)

                    def get_token(self):
                        return self.auth.tenant

                return FakeCatalog(self)

        auth_strategy = FakeAuth

    if auth_strategy:
        kwargs['auth_strategy'] = auth_strategy

    if not user.requirements.is_admin:
        auth_url = CONFIG.trove_auth_url
    else:
        auth_url = CONFIG.values.get('trove_admin_auth_url',
                                     CONFIG.trove_auth_url)

    if CONFIG.values.get('trove_client_cls'):
        cls_name = CONFIG.trove_client_cls
        kwargs['client_cls'] = import_class(cls_name)

    kwargs['tenant'] = user.tenant
    kwargs['auth_url'] = auth_url
    return (user.auth_user, user.auth_key), kwargs


def create_client(cls, user):
    args, kwargs = create_client_args(user)
    kwargs['client_cls'] = cls
    client = Dbaas(*args, **kwargs)
    return client


def make_client(user):
    args, kwargs = create_client_args(user)
    kwargs['client_cls'] = JsonClient
    client = Dbaas(*args, **kwargs)
    client.client.name = "auth"
    client.authenticate()
    return client


def write_snippet(get_replace_list, client, name, url, method, status, reason,
                  func, *func_args):
    """
    'name' is the name of the file, while 'url,' 'method,' 'status,'
    and 'reason' are expected values that are asserted against.
    If func_args is present, it is a list of lists, each one of which
    is passed as the *args to the two invocations of "func".
    """
    func_args = func_args or []
    snippet_writer = SnippetWriter(conf, get_replace_list)
    results = []
    client.client.snippet_writer = snippet_writer
    client.client.name = name
    args = func_args
    result = func(client, *args)

    # Now write the snippet (if this happens earlier we can't replace
    # data such as the instance ID).
    client.client.write_snippet()
    with Check() as check:
        check.equal(client.client.old_info['url'], url)
        check.equal(client.client.old_info['method'], method)
        check.equal(client.client.old_info['response_headers'].status,
                    status)
        check.equal(client.client.old_info['response_headers'].reason,
                    reason)
    results.append(result)
    # To prevent this from writing a snippet somewhere else...
    client.client.name = "junk"

    return results


JSON_INDEX = 0


class Example(object):

    @classmethod
    def get_replace_list(cls):
        return []

    def snippet(self, *args, **kwargs):
        return write_snippet(self.get_replace_list, self.client,
                             *args, **kwargs)


@test(depends_on=[load_config_file], enabled=False)
class Versions(Example):

    @before_class
    def setup(self):
        self.client = make_client(normal_user)

    @test
    def get_versions(self):
        self.snippet(
            "versions",
            "", "GET", 200, "OK",
            lambda client: client.versions.index(conf['version_url']))

    @test
    def get_version(self):
        def version_call(client):
            return client.versions.index(conf['version_url'] + "/v1.0/")

        self.snippet("versions", "/v1.0", "GET", 200, "OK", get_version)


@test(depends_on=[load_config_file])
class Flavors(Example):

    @before_class
    def setup(self):
        self.client = make_client(normal_user)

    @test
    def get_flavors(self):
        self.snippet(
            "flavors",
            "/flavors", "GET", 200, "OK",
            lambda client: client.flavors.list())

    @test
    def get_flavor_by_id(self):
        self.snippet(
            "flavors_by_id",
            "/flavors/1", "GET", 200, "OK",
            lambda client: client.flavors.get(1))


@test(depends_on=[load_config_file])
def clean_slate():
    client = create_client(TroveHTTPClient, admin_user)
    client.client.name = "list"
    instances = client.instances.list()
    assert_equal(0, len(instances), "Instance count must be zero.")


@test(depends_on=[clean_slate])
class CreateInstance(Example):

    @before_class
    def setup(self):
        self.client = make_client(normal_user)

    @test
    def post_create_instance(self):
        set_fake_stuff(uuid=EXAMPLE_INSTANCE_ID)

        def create_instance(client, name):
            instance = client.instances.create(
                name, 1,
                volume={'size': 2},
                databases=[
                    {
                        "name": "sampledb",
                        "character_set": "utf8",
                        "collate": "utf8_general_ci"
                    },
                    {
                        "name": "nextround"
                    }
                ],
                users=[
                    {
                        "databases": [{"name": "sampledb"}],
                        "name": "demouser",
                        "password": "demopassword"
                    }
                ])
            assert_equal(instance.status, "BUILD")
            return instance
        self.instances = self.snippet(
            "create_instance",
            "/instances", "POST", 200, "OK",
            create_instance,
            "json_rack_instance")

    def an_instance_is_not_active(self):
        for instance in self.instances:
            instance = self.client.instances.get(instance.id)
            if instance.status not in CONFIG.running_status:
                assert_equal(instance.status, "BUILD")
                return True
        return False

    @test(depends_on=[post_create_instance])
    def wait_for_instances(self):
        while self.an_instance_is_not_active():
            time.sleep(1)
        global json_instance
        json_instance = self.instances[0]


@test(depends_on=[CreateInstance], groups=['uses_instances'])
class Databases(Example):

    @before_class
    def setup(self):
        self.client = make_client(normal_user)

    @test
    def post_create_databases(self):
        self.snippet(
            "create_databases",
            "/instances/%s/databases" % json_instance.id,
            "POST", 202, "Accepted",
            lambda client: client.databases.create(
                json_instance.id,
                databases=[
                    {
                        "name": "testingdb",
                        "character_set": "utf8",
                        "collate": "utf8_general_ci"
                    }, {
                        "name": "anotherdb"
                    }, {
                        "name": "oneMoreDB"
                    }]))

    @test(depends_on=[post_create_databases])
    def get_list_databases(self):
        self.snippet(
            "list_databases",
            "/instances/%s/databases" % json_instance.id,
            "GET", 200, "OK",
            lambda client: client.databases.list(json_instance.id))

    @test(depends_on=[post_create_databases])
    def get_list_databases_limit_two(self):
        results = self.snippet(
            "list_databases_pagination",
            "/instances/%s/databases?limit=1" % json_instance.id,
            "GET", 200, "OK",
            lambda client: client.databases.list(json_instance.id, limit=1))
        assert_equal(1, len(results[JSON_INDEX]))
        assert_equal("anotherdb", results[JSON_INDEX].next)

    @test(depends_on=[post_create_databases],
          runs_after=[get_list_databases, get_list_databases_limit_two])
    def delete_databases(self):
        self.snippet(
            "delete_databases",
            "/instances/%s/databases/testingdb" % json_instance.id,
            "DELETE", 202, "Accepted",
            lambda client:
            client.databases.delete(json_instance.id, 'testingdb'))


@test(depends_on=[CreateInstance], groups=['uses_instances'])
class Users(Example):

    @before_class
    def setup(self):
        self.client = make_client(normal_user)

    @test
    def post_create_users(self):
        self.snippet(
            "create_users",
            "/instances/%s/users" % json_instance.id,
            "POST", 202, "Accepted",
            lambda client: client.users.create(
                json_instance.id,
                [{
                    "name": "dbuser1",
                    "password": "password",
                    "databases": [
                        {
                            "name": "databaseA"
                        }
                    ]
                }, {
                    "name": "dbuser2",
                    "password": "password",
                    "databases": [
                        {
                            "name": "databaseB"
                        },
                        {
                            "name": "databaseC"
                        }
                    ]
                }, {
                    "name": "dbuser3",
                    "password": "password",
                    "databases": [
                        {
                            "name": "databaseD"
                        }
                    ]
                }]))

    @test(depends_on=[post_create_users])
    def get_list_users(self):
        self.snippet(
            "list_users",
            "/instances/%s/users" % json_instance.id,
            "GET", 200, "OK",
            lambda client: client.users.list(json_instance.id))

    @test(depends_on=[post_create_users])
    def get_list_users_limit_two(self):
        self.snippet(
            "list_users_pagination",
            "/instances/%s/users?limit=2" % json_instance.id,
            "GET", 200, "OK",
            lambda client: client.users.list(json_instance.id, limit=2))

    @test(depends_on=[post_create_users],
          runs_after=[get_list_users, get_list_users_limit_two])
    def delete_users(self):
        user_name = "demouser"
        self.snippet(
            "delete_users",
            "/instances/%s/users/%s" % (json_instance.id, user_name),
            "DELETE", 202, "Accepted",
            lambda client: client.users.delete(json_instance.id,
                                               username=user_name))

    @test(depends_on=[post_create_users])
    def modify_user_attributes(self):
        old_user_name = "dbuser1"
        self.snippet(
            "change_user_attributes",
            "/instances/%s/users/%s" % (json_instance.id, old_user_name),
            "PUT", 202, "Accepted",
            lambda client: client.users.update_attributes(
                json_instance.id,
                username=old_user_name,
                newuserattr={
                    "name": "new_username",
                    "password": "new_password"
                }
            )
        )


@test(depends_on=[CreateInstance], groups=['uses_instances'])
class Root(Example):

    @before_class
    def setup(self):
        self.client = make_client(normal_user)

    @test
    def post_enable_root_access(self):
        self.snippet(
            "enable_root_user",
            "/instances/%s/root" % json_instance.id,
            "POST", 200, "OK",
            lambda client: client.root.create(json_instance.id))

    @test(depends_on=[post_enable_root_access])
    def get_check_root_access(self):
        results = self.snippet(
            "check_root_user",
            "/instances/%s/root" % json_instance.id,
            "GET", 200, "OK",
            lambda client: client.root.is_root_enabled(json_instance.id))
        assert_equal(results[JSON_INDEX].rootEnabled, True)

    @test(depends_on=[get_check_root_access])
    def delete_disable_root_access(self):
        self.snippet(
            "disable_root_user",
            "/instances/%s/root" % json_instance.id,
            "DELETE", 204, "No Content",
            lambda client: client.root.delete(json_instance.id))

        # restore root for subsequent tests
        self.post_enable_root_access()


class ActiveMixin(Example):
    """Adds a method to wait for instance status to become ACTIVE."""

    def _wait_for_active(self, *acceptable_states):
        global json_instance
        json_instance = self.client.instances.get(json_instance.id)
        while json_instance.status not in CONFIG.running_status:
            assert_true(
                json_instance.status in acceptable_states,
                "Instance status == %s; expected it to be one of: %s"
                % (json_instance.status, acceptable_states))
            time.sleep(0.1)
            json_instance = self.client.instances.get(json_instance.id)

    def _wait_for_restore_active(self, *acceptable_states):
        for instance in (self.json_restore, ):
            instance = self.client.instances.get(instance.id)
            while instance.status not in CONFIG.running_status:
                assert_true(
                    instance.status in acceptable_states,
                    "Instance status == %s; expected it to be one of: %s"
                    % (instance.status, acceptable_states))
                time.sleep(0.1)
                instance = self.client.instances.get(instance.id)


STATE = {
    "CONFIGURATION": None,
    "DATASTORE_ID": None,
    "DATASTORE_VERSION_ID": None,
}


@test(depends_on=[CreateInstance], groups=['uses_instances'])
class Datastores(Example):

    @before_class
    def setup(self):
        self.client = make_client(normal_user)

    @test
    def get_datastores_list(self):
        self.datastores = self.snippet(
            "datastores_list",
            "/datastores",
            "GET", 200, "OK",
            lambda client: client.datastores.list())
        for result in self.datastores:
            assert_equal(1, len(result))

    @test(depends_on=[get_datastores_list])
    def get_datastore_by_id(self):
        ds, = self.datastores
        mysql_ds = [x for x in ds if x.name == 'mysql']
        if not mysql_ds:
            fail('no mysql datastore found in list')
        ds_id = STATE["DATASTORE_ID"] = mysql_ds[JSON_INDEX].id
        self.datastore = self.snippet(
            "datastore_by_id",
            "/datastores/%s" % ds_id,
            "GET", 200, "OK",
            lambda client: client.datastores.get(ds_id))

    @test(depends_on=[get_datastore_by_id])
    def get_datastore_versions_list(self):
        ds_id = STATE["DATASTORE_ID"]
        self.datastore_versions = self.snippet(
            "datastore_versions_list",
            "/datastores/%s/versions" % ds_id,
            "GET", 200, "OK",
            lambda client: client.datastore_versions.list(ds_id))

    @test(depends_on=[get_datastore_versions_list])
    def get_datastore_version_by_id(self):
        ds_id = STATE["DATASTORE_ID"]
        ds_v_id = STATE["DATASTORE_VERSION_ID"] = (
            self.datastore_versions[JSON_INDEX][0].id
        )
        self.datastore_version = self.snippet(
            "datastore_version_by_id",
            "/datastores/%s/versions/%s" % (ds_id, ds_v_id),
            "GET", 200, "OK",
            lambda client: client.datastore_versions.get(ds_id, ds_v_id))


@test(depends_on=[Datastores], groups=['uses_instances'])
class Configurations(ActiveMixin):

    @before_class
    def setup(self):
        self.client = make_client(normal_user)

    @test
    def get_configuration_parameters_for_datastore_version(self):
        ds_id = STATE["DATASTORE_ID"]
        ds_v_id = STATE["DATASTORE_VERSION_ID"]
        self.snippet(
            "configuration_parameters_for_datastore_version",
            "/datastores/%s/versions/%s/parameters" % (ds_id, ds_v_id),
            "GET", 200, "OK",
            lambda client: client.configuration_parameters.parameters(
                ds_id, ds_v_id
            )
        )

    @test
    def get_configuration_parameters_without_datastore_version(self):
        ds_v_id = STATE["DATASTORE_VERSION_ID"]
        self.params = self.snippet(
            "configuration_parameters_without_datastore_version",
            "/datastores/versions/%s/parameters" % (ds_v_id),
            "GET", 200, "OK",
            lambda client: (
                client.configuration_parameters.parameters_by_version(ds_v_id)
            )
        )
        assert_true(self.params)

    @test(depends_on=[get_configuration_parameters_without_datastore_version])
    def get_configuration_parameter_for_datastore_version(self):
        ds_id = STATE["DATASTORE_ID"]
        ds_v_id = STATE["DATASTORE_VERSION_ID"]
        param = self.params[JSON_INDEX][0].name
        self.snippet(
            "configuration_parameter_for_datastore_version",
            "/datastores/%s/versions/%s/parameters/%s"
            % (ds_id, ds_v_id, param),
            "GET", 200, "OK",
            lambda client: client.configuration_parameters.get_parameter(
                ds_id, ds_v_id, param))

    @test(depends_on=[get_configuration_parameters_without_datastore_version])
    def get_configuration_parameter_without_datastore_version(self):
        ds_v_id = STATE["DATASTORE_VERSION_ID"]
        param = self.params[JSON_INDEX][0].name

        def get_param(client):
            return client.configuration_parameters.get_parameter_by_version(
                ds_v_id,
                param
            )

        self.params = self.snippet(
            "configuration_parameter_without_datastore_version",
            "/datastores/versions/%s/parameters/%s" % (ds_v_id, param),
            "GET", 200, "OK",
            get_param
        )

    @test(depends_on=[get_configuration_parameter_without_datastore_version])
    def create_configuration(self):
        set_fake_stuff(uuid=EXAMPLE_CONFIG_ID)
        ds_id = STATE["DATASTORE_ID"]
        ds_v_id = STATE["DATASTORE_VERSION_ID"]
        values = {
            "connect_timeout": 120,
            "collation_server": "latin1_swedish_ci"
        }

        def create(client):
            config = client.configurations.create(
                'example-configuration-name', json.dumps(values),
                'example description', ds_id, ds_v_id)
            return config

        self.configurations = self.snippet(
            "configuration_create",
            "/configurations",
            "POST", 200, "OK",
            create)
        STATE["CONFIGURATION"] = self.configurations[JSON_INDEX]

    @test(depends_on=[create_configuration])
    def get_configuration(self):
        config = STATE["CONFIGURATION"]
        self.config = self.snippet(
            "configuration_details",
            "/configurations/%s" % config.id,
            "GET", 200, "OK",
            lambda client: client.configurations.get(config.id))

    @test(depends_on=[create_configuration])
    def list_configurations(self):
        self.configs = self.snippet(
            "configuration_list",
            "/configurations",
            "GET", 200, "OK",
            lambda client: client.configurations.list())

    @test(depends_on=[list_configurations, get_configuration])
    def edit_configuration(self):
        config = STATE["CONFIGURATION"]
        values = {
            'connect_timeout': 300
        }
        self.snippet(
            "configuration_edit_parameters",
            "/configurations/%s" % config.id,
            "PATCH", 200, "OK",
            lambda client: client.configurations.edit(
                config.id, json.dumps(values)))

    @test(depends_on=[edit_configuration])
    def update_configuration(self):
        config = STATE["CONFIGURATION"]
        values = {
            'connect_timeout': 150,
            'collation_server': 'utf8_unicode_ci'
        }
        self.snippet(
            "configuration_update_parameters",
            "/configurations/%s" % config.id,
            "PUT", 202, "Accepted",
            lambda client: client.configurations.update(
                config.id, json.dumps(values),
                'example-updated-name', 'example updated description'))

    @test(depends_on=[update_configuration])
    def attach_configuration_to_instance(self):
        config = STATE["CONFIGURATION"]
        self.snippet(
            "configuration_attach_to_instance",
            "/instances/%s" % json_instance.id,
            "PUT", 202, "Accepted",
            lambda client: client.instances.modify(
                json_instance.id,
                config.id
            )
        )

    @test(depends_on=[attach_configuration_to_instance])
    def list_configurations_instances(self):
        config = STATE["CONFIGURATION"]
        self.config_instances = self.snippet(
            "configuration_list_instances",
            "/configurations/%s/instances" % config.id,
            "GET", 200, "OK",
            lambda client: client.configurations.instances(config.id))

    @test(depends_on=[list_configurations_instances])
    def detach_configuration_from_instance(self):
        self.snippet(
            "configuration_detach_from_instance",
            "/instances/%s" % json_instance.id,
            "PUT", 202, "Accepted",
            lambda client: client.instances.modify(
                json_instance.id, ""))

    @test(depends_on=[detach_configuration_from_instance])
    def instance_restart_after_configration_change(self):
        self.client.instances.restart(json_instance.id)
        self._wait_for_active("REBOOT")


@test(depends_on=[CreateInstance], groups=['uses_instances'])
class InstanceList(Example):

    @before_class
    def setup(self):
        self.client = make_client(normal_user)

    @test
    def get_list_instance_index(self):
        results = self.snippet(
            "instances_index",
            "/instances", "GET", 200, "OK",
            lambda client: client.instances.list())
        for result in results:
            assert_equal(1, len(result))

    @test
    def get_instance_details(self):
        results = self.snippet(
            "instance_status_detail",
            "/instances/%s" % json_instance.id,
            "GET", 200, "OK",
            lambda client: client.instances.get(json_instance.id))
        assert_equal(results[JSON_INDEX].id, json_instance.id)

    @test
    def get_default_instance_configuration(self):
        set_fake_stuff(unique_id=EXAMPLE_CONFIG_SERVER_ID)
        self.snippet(
            "get_default_instance_configuration",
            "/instances/%s/configuration" % json_instance.id,
            "GET", 200, "OK",
            lambda client: client.instances.configuration(json_instance.id))

    @test
    def get_list_instance_index_limit_two(self):
        third_instance = self.client.instances.create(
            "The Third Instance", 1, volume={'size': 2})
        third_instance = self.client.instances.get(third_instance.id)
        while third_instance.status not in CONFIG.running_status:
            time.sleep(0.1)
            third_instance = self.client.instances.get(third_instance.id)

        results = self.snippet(
            "instances_index_pagination",
            "/instances?limit=2", "GET", 200, "OK",
            lambda client: client.instances.list(limit=2))
        for result in results:
            assert_equal(2, len(result))

        self.client.instances.delete(third_instance.id)


@test(depends_on=[CreateInstance], groups=['uses_instances'])
class Backups(ActiveMixin):

    @before_class
    def setup(self):
        self.client = make_client(normal_user)

    @test
    def create_backup(self):
        set_fake_stuff(uuid=EXAMPLE_BACKUP_ID)
        results = self.snippet(
            "backup_create", "/backups", "POST", 202, "Accepted",
            lambda client: client.backups.create(
                name='snapshot',
                instance=json_instance.id,
                description="My Backup"
            )
        )
        self._wait_for_active("BACKUP")
        assert_equal(len(results), 1)
        self.json_backup = results[JSON_INDEX]

    @test(depends_on=[create_backup])
    def create_incremental_backup(self):
        set_fake_stuff(uuid=EXAMPLE_BACKUP_INCREMENTAL_ID)
        results = self.snippet(
            "backup_create_incremental", "/backups", "POST", 202, "Accepted",
            lambda client: client.backups.create(
                name='Incremental Snapshot',
                instance=json_instance.id,
                parent_id=EXAMPLE_BACKUP_ID,
                description="My Incremental Backup"
            )
        )

        self._wait_for_active("BACKUP")
        assert_equal(len(results), 1)
        self.json_backup2 = results[JSON_INDEX]

    @test(depends_on=[create_incremental_backup])
    def get_backup(self):
        results = self.snippet(
            "backup_get",
            "/backups/%s" % self.json_backup.id,
            "GET", 200, "OK",
            lambda client: client.backups.get(self.json_backup.id))
        assert_equal(len(results), 1)

    @test(depends_on=[create_incremental_backup])
    def get_backups_for_instance(self):
        results = self.snippet(
            "backups_by_instance",
            "/instances/%s/backups" % json_instance.id,
            "GET", 200, "OK",
            lambda client: client.instances.backups(json_instance.id))
        assert_equal(len(results), 1)

    @test(depends_on=[create_incremental_backup])
    def list_backups(self):
        results = self.snippet(
            "backup_list",
            "/backups", "GET", 200, "OK",
            lambda client: client.backups.list())
        assert_equal(len(results), 1)

    @test(depends_on=[create_backup])
    def restore(self):
        set_fake_stuff(uuid=EXAMPLE_INSTANCE_ID_2)

        def create_instance(client, name, backup):
            instance = client.instances.create(
                name, 1,
                volume={'size': 2},
                restorePoint={'backupRef': backup})
            assert_equal(instance.status, "BUILD")
            return instance
        results = self.snippet(
            "backup_restore",
            "/instances", "POST", 200, "OK",
            lambda client: create_instance(
                client, "backup_instance", self.json_backup.id))
        assert_equal(len(results), 1)
        self.json_restore = results[JSON_INDEX]
        self._wait_for_restore_active("BUILD")
        self.json_restore = self.client.instances.get(self.json_restore.id)
        asserts.assert_true(self.json_restore.status in CONFIG.running_status)

    @test(depends_on=[restore])
    def delete_restores(self):
        self.snippet(
            "restore_delete",
            "/instances/%s" % self.json_restore.id,
            "DELETE", 202, "Accepted",
            lambda client: client.instances.delete(self.json_restore.id))
        self.json_restore = self.client.instances.get(self.json_restore.id)
        assert_equal(self.json_restore.status, "SHUTDOWN")

    @test(depends_on=[create_backup],
          runs_after=[get_backup, list_backups, restore,
                      get_backups_for_instance])
    def delete_backup(self):
        results = self.snippet(
            "backup_delete",
            "/backups/%s" % self.json_backup.id,
            "DELETE", 202, "Accepted",
            lambda client: client.backups.delete(self.json_backup.id))
        assert_equal(len(results), 1)


@test(depends_on=[CreateInstance], groups=['uses_instances'])
class Actions(ActiveMixin):

    @before_class
    def setup(self):
        self.client = make_client(normal_user)

    @test
    def instance_restart(self):
        self.snippet(
            "instance_restart",
            "/instances/%s/action" % json_instance.id,
            "POST", 202, "Accepted",
            lambda client: client.instances.restart(json_instance.id))
        self._wait_for_active("REBOOT")

    @test
    def instance_resize_volume(self):
        self.snippet(
            "instance_resize_volume",
            "/instances/%s/action" % json_instance.id,
            "POST", 202, "Accepted",
            lambda client: client.instances.resize_volume(json_instance.id, 4))
        self._wait_for_active("RESIZE")
        assert_equal(json_instance.volume['size'], 4)

    @test
    def instance_resize_flavor(self):
        self.snippet(
            "instance_resize_flavor",
            ("/instances/%s/action" % json_instance.id),
            "POST", 202, "Accepted",
            lambda client: client.instances.resize_instance(
                json_instance.id, 3))
        self._wait_for_active("RESIZE")
        # TODO(imsplitbit): remove coercion when troveclient fixes are in
        assert_equal(int(json_instance.flavor['id']), 3)


@test(depends_on=[CreateInstance], groups=['uses_instances', "MgmtHosts"])
class MgmtHosts(Example):

    @before_class
    def setup(self):
        self.client = make_client(admin_user)

    @test
    def mgmt_list_hosts(self):
        results = self.snippet(
            "mgmt_list_hosts",
            "/mgmt/hosts", "GET", 200, "OK",
            lambda client: client.mgmt.hosts.index())

        with Check() as check:
            for hosts in results:
                check.equal(2, len(hosts))
                check.true("fake_host_1" == hosts[0].name
                           or "fake_host_1" == hosts[1].name)
                check.true("fake_host_2" == hosts[0].name
                           or "fake_host_2" == hosts[1].name)
                check.true(1 == results[0][1].instanceCount
                           or 1 == results[0][0].instanceCount)

    @test
    def mgmt_get_host_detail(self):
        results = self.snippet(
            "mgmt_get_host_detail",
            "/mgmt/hosts/fake_host_1", "GET", 200, "OK",
            lambda client: client.mgmt.hosts.get("fake_host_1"))
        with Check() as check:
            for host in results:
                check.equal(results[0].name, "fake_host_1")
                # XML entries won't come back as these types. :(
                check.true(isinstance(results[0].percentUsed, int)),
                check.true(isinstance(results[0].totalRAM, int)),
                check.true(isinstance(results[0].usedRAM, int)),
        with Check() as check:
            for host in results:
                check.equal(1, len(host.instances))
                for instance in host.instances:
                    check.equal(instance['status'], 'HEALTHY')
                    check.true(isinstance(instance['name'], six.string_types))
                    check.true(isinstance(instance['id'], six.string_types))
                    check.true(isinstance(instance['server_id'],
                               six.string_types))
                    check.true(isinstance(instance['tenant_id'],
                               six.string_types))

    @test
    def mgmt_host_update_all(self):
        raise SkipTest("This isn't working... :(")
        self.snippet(
            "mgmt_host_update",
            "/mgmt/hosts/fake_host_1/instances/action",
            "POST", 202, "Accepted",
            lambda client: client.mgmt.hosts.update_all("fake_host_1"))


@test(depends_on=[CreateInstance], groups=['uses_instances'])
class MgmtStorage(Example):

    @before_class
    def setup(self):
        self.client = make_client(admin_user)

    @test
    def mgmt_get_storage(self):
        results = self.snippet(
            "mgmt_get_storage",
            "/mgmt/storage", "GET", 200, "OK",
            lambda client: client.mgmt.storage.index())
        for index, devices in enumerate(results):
            with Check() as check:
                check.equal(1, len(devices))
                device = devices[0]
                check.equal(int(device.capacity['available']), 90)
                check.equal(int(device.capacity['total']), 100)
                check.equal(device.name, "fake_storage")
                check.equal(int(device.provision['available']), 40)
                check.equal(int(device.provision['percent']), 10)
                check.equal(int(device.provision['total']), 50)
                check.equal(device.type, "test_type")
                check.equal(int(device.used), 10)
                if index == JSON_INDEX:
                    check.true(isinstance(device.capacity['available'], int))
                    check.true(isinstance(device.capacity['total'], int))
                    check.true(isinstance(device.provision['available'], int))
                    check.true(isinstance(device.provision['percent'], int))
                    check.true(isinstance(device.provision['total'], int))
                    check.true(isinstance(device.used, int))


@test(depends_on=[CreateInstance], groups=['uses_instances'])
class MgmtAccount(Example):

    @before_class
    def setup(self):
        self.client = make_client(admin_user)

    @test
    def mgmt_get_account_details(self):
        results = self.snippet(
            "mgmt_get_account_details",
            "/mgmt/accounts/%s" % conf['normal_user_tenant'],
            "GET", 200, "OK",
            lambda client: client.mgmt.accounts.show(
                conf['normal_user_tenant'], ))
        with Check() as check:
            for account_info in results:
                check.equal(conf['normal_user_tenant'], account_info.id)

    @test
    def mgmt_get_account_list(self):
        results = self.snippet(
            "mgmt_list_accounts",
            "/mgmt/accounts", "GET", 200, "OK",
            lambda client: client.mgmt.accounts.index())
        matches = {conf['normal_user_tenant']: 2,
                   conf['admin_user_tenant']: 0}
        for index, result in enumerate(results):
            for account in result.accounts:
                if account['id'] not in matches:
                    fail("Did not expect this account ID: %s" % account['id'])
                expected_count = matches[account['id']]
                if index == JSON_INDEX:
                    assert_equal(2, expected_count)
                else:
                    assert_equal(2, expected_count)


def for_both(func):
    @functools.wraps(func)
    def both(self):
        for result in self.results:
            func(self, result)
    return both


@test(depends_on=[CreateInstance], groups=['uses_instances'])
class MgmtInstance(Example):

    @before_class
    def mgmt_get_instance_details(self):
        self.client = make_client(admin_user)
        self.results = self.snippet(
            "mgmt_get_instance_details",
            ("/mgmt/instances/%s" % json_instance.id),
            "GET", 200, "OK",
            lambda client: client.mgmt.instances.show(json_instance.id))

    @test
    @for_both
    def created(self, result):
        assert_true(isinstance(result.created, six.string_types))

    @test
    def deleted(self):
        assert_equal(self.results[JSON_INDEX].deleted, False)

    @test
    @for_both
    def flavor(self, result):
        # TODO(imsplitbit): remove the coercion when python-troveclient fixes
        # land in the public.
        assert_true(
            int(result.flavor['id']) == 1 or int(result.flavor['id']) == 3)
        assert_equal(len(result.flavor['links']), 2)

    @test
    @for_both
    def guest_status(self, result):
        assert_equal(result.guest_status['state_description'], 'running')

    @test(enabled=False)
    @for_both
    def host(self, result):
        assert_equal(result.host, 'fake_host_1')

    @test
    def id(self):
        assert_equal(self.results[JSON_INDEX].id, json_instance.id)

    @test
    @for_both
    def links(self, result):
        assert_true(isinstance(result.links, list))
        for link in result.links:
            assert_true(isinstance(link, dict))
            assert_true(isinstance(link['href'], six.string_types))
            assert_true(isinstance(link['rel'], six.string_types))

    @test
    def local_id(self):
        assert_true(isinstance(self.results[JSON_INDEX].server['local_id'],
                    int))

    @test
    @for_both
    def name(self, result):
        assert_true(isinstance(result.name, six.string_types))

    @test
    @for_both
    def server_id(self, result):
        assert_true(isinstance(result.server['id'], six.string_types))

    @test
    @for_both
    def status(self, result):
        assert_equal("ACTIVE", result.status)

    @test
    @for_both
    def task_description(self, result):
        assert_equal(result.task_description, "No tasks for the instance.")

    @test
    @for_both
    def tenant_id(self, result):
        assert_equal(result.tenant_id, conf['normal_user_tenant'])

    @test
    @for_both
    def updated(self, result):
        assert_true(isinstance(result.updated, six.string_types))

    @test
    @for_both
    def volume(self, result):
        assert_true(isinstance(result.volume, dict))
        assert_true('id' in result.volume)
        assert_true('size' in result.volume)


@test(depends_on=[CreateInstance], groups=['uses_instances'])
class MgmtInstanceIndex(Example):

    @before_class
    def setup(self):
        self.client = make_client(admin_user)

    @test
    def mgmt_instance_index(self, deleted=False):
        self.snippet(
            "mgmt_instance_index",
            "/mgmt/instances?deleted=false", "GET", 200, "OK",
            lambda client: client.mgmt.instances.index(deleted=False))


@test(depends_on=[CreateInstance], groups=['uses_instances'])
class MgmtInstanceDiagnostics(Example):

    @before_class
    def setup(self):
        self.client = make_client(admin_user)

    @test
    def mgmt_get_instance_diagnostics(self):
        self.snippet(
            "mgmt_instance_diagnostics",
            ("/mgmt/instances/%s/diagnostics" % json_instance.id),
            "GET", 200, "OK",
            lambda client: client.diagnostics.get(json_instance.id))


@test(depends_on=[CreateInstance])
class MgmtInstanceRoot(Example):

    @before_class
    def setup(self):
        self.client = make_client(admin_user)

    @test
    def mgmt_get_root_details(self):
        self.snippet(
            "mgmt_get_root_details",
            ("/mgmt/instances/%s/root" % json_instance.id),
            "GET", 200, "OK",
            lambda client: client.mgmt.instances.root_enabled_history(
                json_instance.id)
        )


@test(depends_on=[CreateInstance], enabled=False)
class MgmtInstanceHWInfo(Example):

    @before_class
    def setup(self):
        self.client = make_client(admin_user)

    @test
    def mgmt_get_hw_info(self):
        self.snippet(
            "mgmt_get_hw_info",
            ("/mgmt/instances/%s/hwinfo" % json_instance.id),
            "GET", 200, "OK",
            lambda client, id: client.hw_info.get(id),
            ([json_instance.id], ))


@test(depends_on=[CreateInstance], groups=['uses_instances'])
class MgmtInstanceReboot(Example):

    @before_class
    def setup(self):
        self.client = make_client(admin_user)

    @test
    def mgmt_instance_reboot(self):
        self.snippet(
            "instance_reboot",
            ("/mgmt/instances/%s/action" % json_instance.id),
            "POST", 202, "Accepted",
            lambda client: client.mgmt.instances.reboot(json_instance.id))


@test(depends_on=[CreateInstance],
      groups=['uses_instances'], enabled=False)
class MgmtInstanceGuestUpdate(Example):

    @before_class
    def setup(self):
        self.client = make_client(admin_user)

    @test
    def mgmt_instance_guest_update(self):
        self.snippet(
            "guest_update",
            ("/mgmt/instances/%s/action" % json_instance.id),
            "POST", 202, "Accepted",
            lambda client: client.mgmt.instances.update(json_instance.id))


@test(depends_on=[CreateInstance], runs_after_groups=['uses_instances'])
class ZzzDeleteInstance(Example):

    @before_class
    def setup(self):
        self.client = make_client(normal_user)

    @test
    def zzz_delete_instance(self):
        global json_instance
        self.snippet(
            "delete_instance",
            "/instances/%s" % json_instance.id,
            "DELETE", 202, "Accepted",
            lambda client: client.instances.delete(json_instance.id))
        json_instance = self.client.instances.get(json_instance.id)
        assert_equal(json_instance.status, "SHUTDOWN")

    @test(depends_on=[zzz_delete_instance])
    def delete_configuration(self):
        config = STATE["CONFIGURATION"]
        self.configs = self.snippet(
            "configuration_delete",
            ("/configurations/%s" % config.id),
            "DELETE", 202, "Accepted",
            lambda client: client.configurations.delete(config.id))


if __name__ == "__main__":
    CONFIG.load_from_file("etc/tests/localhost.test.conf")
    TestProgram().run_and_exit()



# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging
from oslo_utils import timeutils

from trove.backup.state import BackupState
from trove.common import cfg
from trove.common.i18n import _
from trove.common.strategies.storage import get_storage_strategy
from trove.conductor import api as conductor_api
from trove.guestagent.dbaas import get_filesystem_volume_stats
from trove.guestagent.strategies.backup.base import BackupError
from trove.guestagent.strategies.backup.base import UnknownBackupType
from trove.guestagent.strategies.backup import get_backup_strategy
from trove.guestagent.strategies.restore import get_restore_strategy

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

CONFIG_MANAGER = CONF.get('mysql'
                          if not CONF.datastore_manager
                          else CONF.datastore_manager)

STRATEGY = CONFIG_MANAGER.backup_strategy
BACKUP_NAMESPACE = CONFIG_MANAGER.backup_namespace
RESTORE_NAMESPACE = CONFIG_MANAGER.restore_namespace
RUNNER = get_backup_strategy(STRATEGY, BACKUP_NAMESPACE)
EXTRA_OPTS = CONF.backup_runner_options.get(STRATEGY, '')

# Try to get the incremental strategy or return the default 'backup_strategy'
INCREMENTAL = CONFIG_MANAGER.backup_incremental_strategy.get(
    STRATEGY, STRATEGY)

INCREMENTAL_RUNNER = get_backup_strategy(INCREMENTAL, BACKUP_NAMESPACE)


class BackupAgent(object):
    def _get_restore_runner(self, backup_type):
        """Returns the RestoreRunner associated with this backup type."""
        try:
            runner = get_restore_strategy(backup_type, RESTORE_NAMESPACE)
        except ImportError:
            raise UnknownBackupType(_("Unknown Backup type: %(type)s in "
                                      "namespace %(ns)s")
                                    % {"type": backup_type,
                                       "ns": RESTORE_NAMESPACE})
        return runner

    def stream_backup_to_storage(self, context, backup_info, runner, storage,
                                 parent_metadata={}, extra_opts=EXTRA_OPTS):
        backup_id = backup_info['id']
        conductor = conductor_api.API(context)

        # Store the size of the filesystem before the backup.
        mount_point = CONFIG_MANAGER.mount_point
        stats = get_filesystem_volume_stats(mount_point)
        backup_state = {
            'backup_id': backup_id,
            'size': stats.get('used', 0.0),
            'state': BackupState.BUILDING,
        }
        conductor.update_backup(CONF.guest_id,
                                sent=timeutils.utcnow_ts(microsecond=True),
                                **backup_state)
        LOG.debug("Updated state for %s to %s.", backup_id, backup_state)

        try:
            with runner(filename=backup_id, extra_opts=extra_opts,
                        **parent_metadata) as bkup:
                LOG.debug("Starting backup %s.", backup_id)
                meta = {}
                meta['datastore'] = backup_info['datastore']
                meta['datastore_version'] = backup_info['datastore_version']
                success, note, checksum, location = storage.save(
                    bkup.manifest,
                    bkup,
                    metadata=meta)

                backup_state.update({
                    'checksum': checksum,
                    'location': location,
                    'note': note,
                    'success': success,
                    'backup_type': bkup.backup_type,
                })

                LOG.debug("Backup %(backup_id)s completed status: "
                          "%(success)s.", backup_state)
                LOG.debug("Backup %(backup_id)s file swift checksum: "
                          "%(checksum)s.", backup_state)
                LOG.debug("Backup %(backup_id)s location: "
                          "%(location)s.", backup_state)

                if not success:
                    raise BackupError(note)

                backup_state.update({'state': BackupState.COMPLETED})

                return meta

        except Exception:
            LOG.exception(
                "Error saving backup: %(backup_id)s.", backup_state)
            backup_state.update({'state': BackupState.FAILED})
            raise
        finally:
            LOG.info("Completed backup %(backup_id)s.", backup_state)
            conductor.update_backup(CONF.guest_id,
                                    sent=timeutils.utcnow_ts(
                                        microsecond=True),
                                    **backup_state)
            LOG.info("Updated state for %s to %s.", backup_id, backup_state)

    def execute_backup(self, context, backup_info,
                       runner=RUNNER, extra_opts=EXTRA_OPTS,
                       incremental_runner=INCREMENTAL_RUNNER):

        LOG.info("Running backup %(id)s.", backup_info)
        storage = get_storage_strategy(
            CONF.storage_strategy,
            CONF.storage_namespace)(context)

        # Check if this is an incremental backup and grab the parent metadata
        parent_metadata = {}
        if backup_info.get('parent'):
            runner = incremental_runner
            LOG.debug("Using incremental backup runner: %s.", runner.__name__)
            parent = backup_info['parent']
            parent_metadata = storage.load_metadata(parent['location'],
                                                    parent['checksum'])
            # The parent could be another incremental backup so we need to
            # reset the location and checksum to *this* parents info
            parent_metadata.update({
                'parent_location': parent['location'],
                'parent_checksum': parent['checksum']
            })

        self.stream_backup_to_storage(context, backup_info, runner, storage,
                                      parent_metadata, extra_opts)

    def execute_restore(self, context, backup_info, restore_location):
        try:
            restore_runner = self._get_restore_runner(backup_info['type'])
            storage = get_storage_strategy(
                CONF.storage_strategy,
                CONF.storage_namespace)(context)

            runner = restore_runner(storage, location=backup_info['location'],
                                    checksum=backup_info['checksum'],
                                    restore_location=restore_location)
            backup_info['restore_location'] = restore_location

            LOG.info("Restoring instance from backup %(id)s to "
                     "%(restore_location)s", backup_info)

            content_size = runner.restore()
            LOG.info("Restore from backup %(id)s completed successfully "
                     "to %(restore_location)s", backup_info)
            LOG.debug("Restore size: %s", content_size)
        except Exception:
            LOG.exception("Error restoring backup %(id)s", backup_info)
            raise
        else:
            LOG.debug("Restored backup %(id)s", backup_info)



# Copyright 2015 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging

from trove.guestagent.common import operating_system
from trove.guestagent.datastore.galera_common import service as galera_service
from trove.guestagent.datastore.mysql_common import service as mysql_service

LOG = logging.getLogger(__name__)


class MariaDBApp(galera_service.GaleraApp):

    OS = operating_system.get_os()

    def __init__(self, status):
        super(MariaDBApp, self).__init__(
            status, mysql_service.BaseLocalSqlClient,
            mysql_service.BaseKeepAliveConnection)

    @property
    def service_candidates(self):
        service_candidates = super(MariaDBApp, self).service_candidates
        return {
            operating_system.DEBIAN: ["mariadb"] + service_candidates,
            operating_system.REDHAT: ["mariadb"],
            operating_system.SUSE: service_candidates
        }[self.OS]

    @property
    def mysql_service(self):
        result = super(MariaDBApp, self).mysql_service
        if result['type'] == 'sysvinit':
            result['cmd_bootstrap_galera_cluster'] = (
                "sudo service %s bootstrap"
                % result['service'])
        elif result['type'] == 'systemd':
            if operating_system.find_executable('galera_new_cluster'):
                result['cmd_bootstrap_galera_cluster'] = (
                    "sudo galera_new_cluster")
            else:
                result['cmd_bootstrap_galera_cluster'] = (
                    "sudo systemctl start %s@bootstrap.service"
                    % result['service'])
        return result

    @property
    def cluster_configuration(self):
        return self.configuration_manager.get_value('galera')

    def _get_slave_status(self):
        with self.local_sql_client(self.get_engine()) as client:
            return client.execute('SHOW SLAVE STATUS').first()

    def _get_master_UUID(self):
        slave_status = self._get_slave_status()
        return slave_status and slave_status['Master_Server_Id'] or None

    def _get_gtid_executed(self):
        with self.local_sql_client(self.get_engine()) as client:
            return client.execute('SELECT @@global.gtid_binlog_pos').first()[0]

    def get_last_txn(self):
        master_UUID = self._get_master_UUID()
        last_txn_id = '0'
        gtid_executed = self._get_gtid_executed()
        for gtid_set in gtid_executed.split(','):
            uuid_set = gtid_set.split('-')
            if uuid_set[1] == master_UUID:
                last_txn_id = uuid_set[-1]
                break
        return master_UUID, int(last_txn_id)

    def get_latest_txn_id(self):
        LOG.info("Retrieving latest txn id.")
        return self._get_gtid_executed()

    def wait_for_txn(self, txn):
        LOG.info("Waiting on txn '%s'.", txn)
        with self.local_sql_client(self.get_engine()) as client:
            client.execute("SELECT MASTER_GTID_WAIT('%s')" % txn)


class MariaDBRootAccess(mysql_service.BaseMySqlRootAccess):
    def __init__(self):
        super(MariaDBRootAccess, self).__init__(
            mysql_service.BaseLocalSqlClient,
            MariaDBApp(mysql_service.BaseMySqlAppStatus.get()))


class MariaDBAdmin(mysql_service.BaseMySqlAdmin):
    def __init__(self):
        super(MariaDBAdmin, self).__init__(
            mysql_service.BaseLocalSqlClient, MariaDBRootAccess(),
            MariaDBApp)
